# -*- coding: utf-8 -*-
"""
Created on Wed Oct 14 16:32:51 2020

@author: chris.kerklaan

Version 0209.

"""
from osgeo import ogr

TABLES = {
    "all": [
        "v2_simple_infiltration",
        "v2_global_settings",
        "v2_aggregation_settings",
        "v2_interflow",
        "v2_groundwater",
        "v2_1d_boundary_conditions",
        "v2_2d_boundary_conditions",
        "v2_1d_lateral",
        "v2_2d_lateral",
        "v2_calculation_point",
        "v2_connection_nodes",
        "v2_cross_section_location",
        "v2_culvert",
        "v2_channel",
        "v2_connected_pnt",
        "v2_pipe",
        "v2_weir",
        "v2_orifice",
        "v2_manhole",
        "v2_pumpstation",
        "v2_windshielding",
        "v2_cross_section_definition",
        "v2_obstacle",
        "v2_levee",
        "v2_grid_refinement",
        "v2_grid_refinement_area",
        "v2_numerical_settings",
        "v2_dem_average_area",
        "v2_impervious_surface_map",
        "v2_impervious_surface",
        "v2_surface_parameters",
        "v2_surface_map",
        "v2_surface",
        "v2_control",
        "v2_control_delta",
        "v2_control_group",
        "v2_control_measure_group",
        "v2_control_measure_map",
        "v2_control_memory",
        "v2_control_pid",
        "v2_control_table",
        "v2_control_timed",
        "v2_floodfill",
        "v2_pumped_drainage_area",
        "v2_initial_waterlevel",
        "schema_version",
    ],
    "csd": ["v2_pipe", "v2_weir", "v2_culvert"],
    "geom": [
        "v2_connection_nodes",
        "v2_obstacle",
        "v2_culvert",
        "v2_channel",
        "v2_cross_section_location",
    ],
    "order": [
        "v2_pipe",
        "v2_weir",
        "v2_culvert",
        "v2_pumpstation",
        "v2_orifice",
        "v2_cross_section_location",
        "v2_channel",
        "v2_impervious_surface_map",
        "v2_impervious_surface",
        "v2_manhole",
        "v2_1d_boundary_conditions",
        "v2_connection_nodes",
        "v2_cross_section_definition",
        "v2_grid_refinement_area",
        "v2_obstacle",
        "v2_aggregation_settings",
        "v2_global_settings",
        "v2_numerical_settings",
        "v2_simple_infiltration",
    ],
    "single": [
        "v2_manhole",
        "v2_impervious_surface_map",
        "v2_1d_boundary_conditions",
    ],
    "start_end": [
        "v2_weir",
        "v2_pipe",
        "v2_culvert",
        "v2_pumpstation",
        "v2_orifice",
        "v2_channel",
    ],
    "skip": ["v2_aggregation_settings"],
}

raster_fields = [
    "dem_file",
    "frict_coef_file",
    "max_interception_file",
    "interception_file",
    "initial_waterlevel_file",
    "wind_shielding_file",
]

extra_sqlite_tables = [
    "layer_statistics",
    "views_layer_statistics",
    "spatialite_history",
    "views_geometry_columns",
]


connected_tables = {
    "global_settings": ["numerical_settings"],
    "boundary_conditions": [],
    "1d_lateral": ["connection_nodes"],
    "2d_boundary_conditions": [],
    "2d_lateral": [],
    "aggregation_settings": ["global_settings"],
    "channel": ["connection_nodes"],
    "connection_nodes": [],
    "cross_section_definition": [],
    "cross_section_location": ["channel", "cross_section_definition"],
    "culvert": ["cross_section_definition", "connection_nodes"],
    "numercial_settings": [],
    "grid_refinement": [],
    "grid_refinement_area": [],
    "impervious_surface": [],
    "impervious_surface_map": ["connection_nodes", "impervious_surface"],
    "levee": [],
    "manhole": ["connection_nodes"],
    "obstacle": [],
    "orifice": ["cross_section_definition", "connection_nodes"],
    "pipe": ["cross_section_definition", "connection_nodes"],
    "pumpstation": ["connection_nodes"],
    "surface": ["surface_parameters"],
    "surface_map": ["surface", "connection_nodes"],
    "surface_parameters": [],
    "weir": ["cross_section_definition", "connection_nodes"],
    "wind": [],
    "windshielding": ["channel"],
}


connected_tables_reversed = {}
for key1, values in connected_tables.items():
    connected_tables_reversed[key1] = []

    for key2, values2 in connected_tables.items():
        if key2 == key1:
            continue
        else:
            if key1 in values2:
                connected_tables_reversed[key1].append(key2)


class Properties(object):
    def __getitem__(self, key):
        return getattr(self, key)

    def __iter__(self):
        for i in dir(self):
            if "__" not in i:
                yield i, getattr(self, i)

    @property
    def schema_version(self):
        return {
            "version_num": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Version of your schematisation",
                "unit": "id",
            },
        }

    @property
    def v2_1d_boundary_conditions(self):
        return {
            "connected_tables": {},
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
            },
            "boundary_type": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [1, 2, 3, 5],
                "description": """
                                Boundary type.
                                1: waterlevel boundary
                                2: velocity boundary
                                3: discharge boundary 
                                5: Sommerfeld boundary (waterlevel slope). 

                                For types 2, 3 and 5 the channel direction determines sign of the input value. 
                                If the boundary is placed on the channel endpoint, 
                                positive values mean for example water is being extracted from the model.""",
                "unit": "1 M above datum, 2: m/s, 3: m3/s",
            },
            "connection_node_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique connection node id",
                "unit": None,
            },
            "timeseries": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Format has to be 'min,value min,value'",
                "unit": "min,value min,value",
            },
        }

    @property
    def v2_1d_lateral(self):
        return {
            "connected_tables": {"v2_connection_nodes": ["connection_node_id"]},
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
            },
            "connection_node_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique connection node id",
                "unit": None,
            },
            "timeseries": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Lateral for 1D connection nodes",
                "unit": "min, m3/s",
            },
        }

    @property
    def v2_2d_boundary_conditions(self):
        return {
            "connected_tables": {"v2_connection_nodes": ["connection_node_id"]},
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "boundary_type": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [1],
                "description": "Boundary condition for 2d model edge (must be on edge of DEM file)",
                "unit": "1:m",
                "default": None,
            },
            "display_name": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Name field, no constraints",
                "unit": "1:m",
                "default": None,
            },
            "timeseries": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Lateral for 1D connection nodes",
                "unit": "min,value min,value",
                "default": None,
            },
        }

    @property
    def v2_2d_lateral(self):
        return {
            "connected_tables": {},
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "timeseries": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Discharge for lateral",
                "unit": "min, m3/s",
                "default": None,
            },
            "type": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [1],
                "description": "lateral on 2d",
                "unit": None,
                "default": None,
            },
        }

    @property
    def v2_aggregation_settings(self):
        return {
            "connected_tables": {"v2_global_settings": ["global_settings_id"]},
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "aggregation_method": {
                "type": ogr.OFTString,
                "optional": False,
                "options": [
                    "avg",
                    "min",
                    "max",
                    "cum",
                    "med",
                    "cum_negative",
                    "cum_positive",
                    "current",
                ],
                "description": "Method of aggregation, only use 'current' for volume and interception",
                "unit": None,
                "default": None,
            },
            "flow_variable": {
                "type": ogr.OFTString,
                "optional": True,
                "options": [
                    "discharge",
                    "flow_velocity",
                    "pump_discharge",
                    "rain",
                    "waterlevel",
                    "wet_cross-section",
                    "wet_surface",
                    "lateral_discharge",
                    "volume",
                    "simple_infiltration",
                    "leakage",
                    "interception",
                ],
                "description": "The name of output variable that is aggregated.",
                "unit": None,
                "default": None,
            },
            "global_settings_id": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "v2_global_settings scenario id",
                "unit": None,
                "default": None,
            },
            "timestep": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Timestep size for aggregation",
                "unit": "seconds",
                "default": None,
            },
            "var_name": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Method of aggregation, only use 'current' for volume and interception",
                "unit": None,
                "default": None,
            },
        }

    @property
    def v2_channel(self):
        return {
            "connected_tables": {
                "v2_connection_nodes": [
                    "connection_node_end_id",
                    "connection_node_start_id",
                ]
            },
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "calculation_type": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [100, 101, 102, 105],
                "description": """100 = embedded channel
                101 = stand-alone channel
                102 = connected channel
                105 = double connected channel
                Embedded or connected can only be used where a DEM is present. Any start-, end- or calculation node along a channel with these types may not lay outside the DEM.""",
                "unit": None,
                "default": False,
            },
            "code": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Name field, no constraints",
                "unit": None,
                "default": None,
            },
            "connection_node_end_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "End node of channel line, must be present in v2_channel and the channel geometry endpoint must be snappen on the given connection node",
                "unit": None,
                "default": None,
            },
            "connection_node_start_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "End node of channel line, must be present in v2_channel and the channel geometry endpoint must be snappen on the given connection node",
                "unit": None,
                "default": None,
            },
            "display_name": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Name fields, no constraints",
                "unit": None,
                "default": None,
            },
            "dist_calc_points": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "Distance between calculation points on line segments",
                "unit": "Meter",
                "default": None,
            },
            "zoom_category": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [0, 1, 2, 3, 4, 5],
                "description": "Visibility in livesite",
                "unit": None,
                "default": None,
            },
            "the_geom": {
                "type": ogr.wkbLineString,
                "optional": False,
                "options": None,
                "description": "Geometry",
                "unit": "Depend on projection",
                "default": None,
            },
        }

    @property
    def v2_connection_nodes(self):
        return {
            "connected_tables": {},
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "code": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Name field, no constraints",
                "unit": None,
                "default": False,
            },
            "initial_waterlevel": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "Initial water level at connection node",
                "unit": "Meter above datum",
                "default": False,
            },
            "storage_area": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "Storage area, e.g., for manholes in sewerage calculations",
                "unit": "Meter squared",
                "default": None,
            },
            "the_geom": {
                "type": ogr.wkbPoint,
                "optional": False,
                "options": None,
                "description": "Geometry",
                "unit": "Depend on projection",
                "default": None,
            },
        }

    @property
    def v2_cross_section_definition(self):
        return {
            "connected_tables": {},
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "code": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Name field, no constraints",
                "unit": None,
                "default": False,
            },
            "height": {
                "type": ogr.OFTString,
                "optional": True,
                "options": None,
                "description": "For tabulated fill in space-separated heights of profile",
                "unit": "Meter",
                "default": None,
            },
            "shape": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [1, 2, 3, 5, 6],
                "description": """
                1 = rectangle; specify width and height (profile/upper side is not automatically closed)
                2 = circle; specify width (profile/upper side is automatically closed)
                3 = egg; specify only 1 width. From this 3Di creates an egg-shaped profile with heigth = 1.5*width
                5 = tabulated rectangle; specify space-seperated width and height intervals. Between intervals the profile is defined straight. Can be closed by stating width 0 at heighest heigth
                6 = tabulated trapezium; specify space-sperated width and height intervals. Between intervals the profile is interpolated. Can be closed by stating width 0 at heighest heigth""",
                "unit": None,
                "default": None,
            },
            "width": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "For tabulated fill in space-separated widths of profile. Fill in diameter for circle.",
                "unit": "Meter",
                "default": None,
            },
        }

    @property
    def v2_cross_section_location(self):
        return {
            "connected_tables": {
                "v2_channel": ["channel_id"],
                "v2_cross_section_definition": ["definition_id"],
            },
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "bank_level": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "For connected channels only. Reference lavel for exchange between 1D and 2D.",
                "unit": "Meter above datum",
                "default": False,
            },
            "channel_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Reference to channel id.",
                "unit": None,
                "default": None,
            },
            "code": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Name field, no constraints",
                "unit": None,
                "default": None,
            },
            "definition_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Reference to v2_cross_section_definition_id",
                "unit": None,
                "default": None,
            },
            "friction_type": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [1, 2],
                "description": """1 = Chezy !not yet implemented
                                  2 = Manning""",
                "unit": "1: m1/2/s 2: s/m1/3",
                "default": None,
            },
            "friction_value": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Friction or rougness value of profile.",
                "unit": "1: m1/2/s 2: s/m1/3",
                "default": None,
            },
            "reference_level": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Reference level or bottom level of profile",
                "unit": "Meter above datum",
                "default": None,
            },
            "the_geom": {
                "type": ogr.wkbPoint,
                "optional": False,
                "options": None,
                "description": "Geometry",
                "unit": "Depend on projection",
                "default": None,
            },
        }

    @property
    def v2_culvert(self):
        return {
            "connected_tables": {
                "v2_cross_section_definition": ["cross_section_definition_id"],
                "v2_connection_nodes": [
                    "connection_node_start_id",
                    "connection_node_end_id",
                ],
            },
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "calculation_type": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [100, 101, 102, 105],
                "description": """
                100 = embedded channel
                101 = stand-alone channel
                102 = connected channel
                105 = double connected channel
                Embedded or connected can only be used where a DEM is present. Any start-, end- or calculation node along a channel with these types may not lay outside the DEM.""",
                "unit": "Meter above datum",
                "default": False,
            },
            "code": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Name field, no constraints",
                "unit": None,
                "default": None,
            },
            "connection_node_end_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "End node of culvert, must be present in v2_connection_nodes.",
                "unit": None,
                "default": None,
            },
            "connection_node_start_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "End node of culvert, must be present in v2_connection_nodes.",
                "unit": None,
                "default": None,
            },
            "display_name": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Name fields, no constraints",
                "unit": None,
                "default": None,
            },
            "zoom_category": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [0, 1, 2, 3, 4, 5],
                "description": "Visibility in livesite",
                "unit": None,
                "default": None,
            },
            "cross_section_definition_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Reference to v2_cross_section_definition_id",
                "unit": None,
                "default": None,
            },
            "discharge_coefficient_negative": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "Discharge coefficient for negative flow (from end node to start node). Can be set to 0 when closed",
                "unit": None,
                "default": None,
            },
            "discharge_coefficient_positive": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "Discharge coefficient for positive flow (from end node to start node). Can be set to 0 when closed",
                "unit": None,
                "default": None,
            },
            "dist_calc_points": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Distance between calcultion points of linesegments",
                "unit": "Meter",
                "default": None,
            },
            "friction_type": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [1, 2],
                "description": "Friction type 1 = Chezy (not yet implemented) 2 = Manning",
                "unit": None,
                "default": None,
            },
            "friction_value": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Friction or roughness value for profile",
                "unit": "1: m1/2/s 2: s/m1/3",
                "default": None,
            },
            "invert_level_end_point": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Invert level at culvert endpoint",
                "unit": "Meter above datum",
                "default": None,
            },
            "invert_level_start_point": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Invert level at culvert endpoint",
                "unit": "Meter above datum",
                "default": None,
            },
        }

    @property
    def v2_global_settings(self):
        return {
            "connected_tables": {"v2_numerical_settings": ["numerical_settings_id"]},
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "advection_1d": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [0, 1],
                "description": "0: off, 1: on, Use advection in 1D, other schemes 2-7 are in experimental fase.",
                "unit": None,
                "default": 0,
            },
            "advection_2d": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [0, 1],
                "description": "0: off, 1: on, Use advection in 2D.",
                "unit": None,
                "default": 0,
            },
            "control_group_id": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "Reference to id that contains control settings for this scenario.",
                "unit": None,
                "default": None,
            },
            "dem_file": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Relative to dem file (.tif)",
                "unit": "Meter above datum",
                "default": None,
            },
            "dem_obstacle_detection": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "Automatically detect obstacles based on DEM-file.Works only in combination with dem_obstacle_height (has no relation with v2_obstacle).",
                "unit": None,
                "default": None,
            },
            "dem_obstacle_height": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "Relative height (above lowest pixel of calc cell) for obstacle detection.",
                "unit": "Above lowest pixel of calc cell",
                "default": None,
            },
            "dist_calc_points": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Global distance between calculation points for line elements",
                "unit": "Meter",
                "default": None,
            },
            "embedded_cutoff_threshold": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "Relative length of cell size. When embedded channel intersects 2D grid size over length shorter than the cellsize * cutoff threshold, the embedded channel skips this 2D cell. Usefull for preventing very short embedded channel segments (which slow down your model).",
                "unit": "Factor between 0 and 1",
                "default": 0.05,
            },
            "epsg_code": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Define map projection for study area. Must match raster projection.",
                "unit": "EPSG projection",
                "default": 28992,
            },
            "flooding_threshold": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Water depth threshold for flow between 2D cells. Depth relative to lowest bathymetry pixel at the edge between two 2D cell.",
                "unit": "Meter",
                "default": 0.0,
            },
            "frict_avg": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [0, 1],
                "description": "The roughness coefficient will be average within one cell.",
                "unit": None,
                "default": 0,
            },
            "frict_coef": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Constant friction coefficient for 2D.",
                "unit": "1: m1/2/s 2: s/m1/3",
                "default": None,
            },
            "frict_coef_file": {
                "type": ogr.OFTString,
                "optional": True,
                "options": None,
                "description": "Relative path to friction file",
                "unit": "1: m1/2/s 2: s/m1/3",
                "default": None,
            },
            "frict_type": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": [1, 2],
                "description": "1: Chezy for 2D, 2: Manning for 2D.",
                "unit": "1: m1/2/s 2: s/m1/3",
                "default": None,
            },
            "grid_space": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Smallest grid size in quadtree, k=1, must be at least twice the cellsize of the raster",
                "unit": "Meters",
                "default": None,
            },
            "groundwater_settings_id": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "Reference to id that contains groundwater settings for this scenario.",
                "unit": None,
                "default": None,
            },
            "initial_groundwater_level": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "Initial groundwater level",
                "unit": "Meter above datum",
                "default": None,
            },
            "initial_groundwater_level_file": {
                "type": ogr.OFTString,
                "optional": True,
                "options": None,
                "description": "Relative path to groundwater file",
                "unit": "Meter above datum",
                "default": None,
            },
            "initial_groundwater_level_type": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": [0, 1, 2],
                "description": """0=max, 1=min, 2=avg
                                  **Mandatory when using initial water level file.
                                  """,
                "unit": None,
                "default": None,
            },
            "initial_waterlevel": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Global initial waterlevel",
                "unit": "Meter above datum",
                "default": None,
            },
            "initial_waterlevel_file": {
                "type": ogr.OFTString,
                "optional": True,
                "options": None,
                "description": "Relative path to initial waterlevel file file",
                "unit": "Meter above datum",
                "default": None,
            },
            "interflow_settings_id": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "Reference to interflow settings id",
                "unit": "Meter above datum",
                "default": None,
            },
            "interception_global": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "Global value for interception.",
                "unit": "Meter",
                "default": None,
            },
            "interception_file": {
                "type": ogr.OFTString,
                "optional": True,
                "options": None,
                "description": "Relative path to interception file.",
                "unit": "Meter",
                "default": None,
            },
            "kmax": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Maximum multitude of smallest grid size in quadtree starting from grid_space at k=1. Grid size increases according to 2^(k-1) * grid_space.",
                "unit": "Meter",
                "default": None,
            },
            "manhole_storage_area": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": """ Default manhole storage area. This is the surface area that each manhole is given when water reaches above the surface level.
                    **Mandatory when using only 1d flow (no dem) manhole area must be larger than 0 (and an INTEGER)
                    Must be NULL when using only 2d.""",
                "unit": "Meter squared",
                "default": None,
            },
            "max_angle_1d_advection": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "Maximum angle at which advection is considered.",
                "unit": "degrees [0-90]",
                "default": None,
            },
            "maximum_sim_time_step": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": """Maximum timestep during simulation.
                                 ** Mandatory when using timestep plus.""",
                "unit": "Seconds",
                "default": None,
            },
            "minimum_sim_time_step": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "Minimum timestep during simulation, keep it small for stability",
                "unit": "Seconds",
                "default": 0.01,
            },
            "maximum_table_step_size": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "Maximum step size for tables.",
                "unit": "meters",
                "default": None,
            },
            "name": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Name",
                "unit": None,
                "default": None,
            },
            "nr_timesteps": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Maximum nr of timesteps",
                "unit": None,
                "default": None,
            },
            "numerical_settings_id": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "Reference to id that contains numerical settings for this scenario.",
                "unit": None,
                "default": None,
            },
            "output_time_step": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Timestep written in the output file.",
                "unit": "Seconds",
                "default": None,
            },
            "sim_time_step": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Simulation time step.",
                "unit": "Seconds",
                "default": None,
            },
            "simple_infiltration_settings_id": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "Reference to id that contains settings for simple infiltration for this scenario.",
                "unit": "Seconds",
                "default": None,
            },
            "start_date": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Start datum",
                "unit": "Format: 2017-01-01",
                "default": None,
            },
            "start_time": {
                "type": ogr.OFTString,
                "optional": True,
                "options": None,
                "description": "Starttime of simulation.",
                "unit": "Format: 00:00:00",
                "default": None,
            },
            "table_step_size": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "User-defined table stepsize/increment (m).",
                "unit": "Meters",
                "default": None,
            },
            "table_step_size_1d": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "User-defined table stepsize/increment (m) for 1d cross-sections and volumes.  default value = table_step_size.",
                "unit": "Meters",
                "default": None,
            },
            "table_step_size_volume_2d": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "User-defined table stepsize/increment (m) for defining 2D volumes. Can increase speed when this is set larger than  table_step_size. default value = table_step_size.",
                "unit": "Meters",
                "default": None,
            },
            "timestep_plus": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": [0, 1],
                "description": "Allow switching to larger timestep when simulation is steady.",
                "unit": None,
                "default": 0,
            },
            "use_0d_inflow": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Include 0D inflow (NWRW/impervious surfaces) in simulation.",
                "unit": None,
                "default": None,
            },
            "use_1d_flow": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Include 1D flow in simulation.",
                "unit": None,
                "default": None,
            },
            "use_2d_flow": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Include 2D flow in simulation.",
                "unit": None,
                "default": None,
            },
            "use_2d_rain": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Use rainfall via 2D surface for this scenario.",
                "unit": None,
                "default": None,
            },
            "water_level_ini_type": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "Use rainfall via 2D surface for this scenario.",
                "unit": None,
                "default": None,
            },
            "wind_shielding_file": {
                "type": ogr.OFTString,
                "optional": True,
                "options": None,
                "description": "NOT YET IMPLEMENTED",
                "unit": None,
                "default": None,
            },
            "guess_dams": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "NOT YET IMPLEMENTED",
                "unit": None,
                "default": None,
            },
        }

    @property
    def v2_groundwater(self):
        return {}  # TODO

    @property
    def v2_grid_refinement(self):
        return {
            "connected_tables": {},
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "code": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Name field, no constraints",
                "unit": None,
                "default": False,
            },
            "display_name": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Name field, no constraints.",
                "unit": None,
                "default": None,
            },
            "refinement_level": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Local refinement level. Starting from 1. Values above kmax (v2_global_settings) are ignored.",
                "unit": None,
                "default": None,
            },
            "the_geom": {
                "type": ogr.wkbLineString,
                "optional": False,
                "options": None,
                "description": "Geometry",
                "unit": "Depend on projection",
                "default": None,
            },
        }

    @property
    def v2_grid_refinement_area(self):
        return {
            "connected_tables": {},
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "code": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Name field, no constraints",
                "unit": None,
                "default": False,
            },
            "display_name": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Name field, no constraints.",
                "unit": None,
                "default": None,
            },
            "refinement_level": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Local refinement level. Starting from 1. Values above kmax (v2_global_settings) are ignored.",
                "unit": None,
                "default": None,
            },
            "the_geom": {
                "type": ogr.wkbPolygon,
                "optional": False,
                "options": None,
                "description": "Geometry",
                "unit": "Depend on projection",
                "default": None,
            },
        }

    @property
    def v2_impervious_surface(self):
        return {
            "connected_tables": {},
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "area": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "area of surface",
                "unit": None,
                "default": None,
            },
            "display_name": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Name field, no constraints.",
                "unit": None,
                "default": None,
            },
            "code": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Name field, no constraints.",
                "unit": None,
                "default": None,
            },
            "dry_weather_flow": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "Dry weather flow per inhabitant",
                "unit": "L/day per inhabitant",
                "default": None,
            },
            "nr_of_inhabitants": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "Number of inhabitant used for dry wheather flow.",
                "unit": "Nr of inhabitants",
                "default": None,
            },
            "surface_class": {
                "type": ogr.OFTString,
                "optional": False,
                "options": [
                    "impervious paving",
                    "pervious paving",
                    "semi-pervious paving",
                    "unpaved",
                    "building",
                ],
                "description": """gesloten verharding -- impervious paving
                                  open verharding       -- pervious paving
                                  half verhard               -- semi-pervious paving
                                  onverhard                  -- unpaved
                                  pand                         -- building""",
                "unit": None,
                "default": None,
            },
            "surface_sub_class": {
                "type": ogr.OFTString,
                "optional": True,
                "options": None,
                "description": """unkown""",
                "unit": None,
                "default": None,
            },
            "surface_inclination": {
                "type": ogr.OFTString,
                "optional": False,
                "options": ["level", "inclined", "elongated"],
                "description": """vlak              -- level
                                hellend        -- inclined
                                uitgestrekt  -- elongated """,
                "unit": None,
                "default": None,
            },
            "zoom_category": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": [0, 1, 2, 3, 4, 5],
                "description": "Name field, no constraints.",
                "unit": None,
                "default": None,
            },
        }

    @property
    def v2_impervious_surface_map(self):
        return {
            "connected_tables": {
                "v2_connection_nodes": ["connection_node_id"],
                "v2_impervious_surface": ["impervious_surface_id"],
            },
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "connection_node_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "ID of connection node",
                "unit": None,
                "default": None,
            },
            "impervious_surface_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "ID of impervious surface feature",
                "unit": None,
                "default": None,
            },
            "percentage": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Percentage of imperivous surface are places on connection nodes/",
                "unit": None,
                "default": None,
            },
        }

    @property
    def v2_interflow(self):
        return {
            "connected_tables": {},
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "display_name": {
                "type": ogr.OFTString,
                "optional": True,
                "options": None,
                "description": "Name field, no constraints.",
                "unit": None,
                "default": None,
            },
            "hydraulic_conductivity": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": """
                            Global hydraulic conductivity (Darcy)
                            **When interflow_type > 0 then hydraulic_conductivity OR hydraulic_conductivity_file is mandatory""",
                "unit": "m/day",
                "default": None,
            },
            "hydraulic_conductivity_file": {
                "type": ogr.OFTString,
                "optional": True,
                "options": None,
                "description": """
                            Global hydraulic conductivity (Darcy)
                            **When interflow_type > 0 then hydraulic_conductivity OR hydraulic_conductivity_file is mandatory""",
                "unit": "m/day",
                "default": None,
            },
            "impervious_layer_elevation": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": """
                                When using interflow: Depth of interflow layer defined below lowest pixel (so always positive). Imaginary bottom of interflow layer. 
                                For interflow types 1 and 2 it is ignored for the volume in the interflow layer (but still it must be filled in when using interflow). The volume in these types is determined by the porosity and the porosity layer thickness. 
                                For interflow types 3 and 4 it is used to determine the volume in the interflow layer. 
                                In all types the waterlevel in the interflow layer starts at this level. It does not influence flow.
                """,
                "unit": "Meter",
                "default": None,
            },
            "interflow_type": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": """
                                Include interflow in simulation.
                                0: No Interflow
                                1: define 1 porosity value for model. This porsity will be rescaled per pixel (to lowest pixel per cell), so (interflow) volume is the same for each pixel within 1 cell --> define porosity, hydraulic_conductivity, porosity_layer_thickness and impervious_layer_elevation (can be used in combination with groundwater)
                                2: define 1 porosity value for model. This porsity will be rescaled per pixel (to lowest pixel whole model), so (interflow) volume is the same for each pixel in whole model --> define porosity, hydraulic_conductivity, porosity_layer_thickness and impervious_layer_elevation (cannot be used in combination with groundwater)
                                3: define 1 porosity value for model. This porsity will not be rescaled, but each pixel in the model has the same porosity. The (interflow) volume for each pixel depends on the impervious_layer_elevation, which is below lowest pixel of cell --> define porosity, hydraulic_conductivity and impervious_layer_elevation (can be used in combination with groundwater)
                                4: define 1 porosity value for model. This porsity will not be rescaled, but each pixel in the model has the same porosity. The (interflow) volume for each pixel depends on the impervious_layer_elevation, which is below lowest pixel of whole model --> define porosity, hydraulic_conductivity and impervious_layer_elevation (cannot be used in combination with groundwater)
                                
                                * NOT NULL
                                ** do not use interflow in combination with limiter_slope_crossectional_area_2d >0 AND/OR limiter_slope_friction_2d>0
                """,
                "unit": None,
                "default": 0,
            },
            "porosity": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": """
                                Porosity (between 0 and 1) of interflow layer.
                                ** Mandatory when using interflow
                """,
                "unit": None,
                "default": None,
            },
            "porosity_file": {
                "type": ogr.OFTString,
                "optional": True,
                "options": None,
                "description": " Relative path to porosity file",
                "unit": None,
                "default": None,
            },
            "porosity_layer_thickness": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": """Thickness of porosity layer relative to DEM.
                                 **Mandatory for interflow_type 1 and 2.""",
                "unit": "Meter",
                "default": None,
            },
        }

    @property
    def v2_levee(self):
        return {
            "connected_tables": {},
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "code": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Name field, no constraints.",
                "unit": None,
                "default": None,
            },
            "crest_level": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Crest level of levee segment",
                "unit": "Meter above datum",
                "default": None,
            },
            "material": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": [1, 2],
                "description": """** Mandatory when you want to use a levee breach during your calculation 
                                     Material used for breach growth.
                                     1: sand
                                     2: clay""",
                "unit": None,
                "default": None,
            },
            "max_breach_depth": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": """** Mandatory when you want to use a levee breach during your calculation 
                                    Maximum breach depth relative to crest level (thus a positive value must be filled in).""",
                "unit": "Meter below crest level",
                "default": None,
            },
            "the_geom": {
                "type": ogr.wkbLineString,
                "optional": False,
                "options": None,
                "description": "Geometry",
                "unit": "Depend on projection",
                "default": None,
            },
        }

    @property
    def v2_manhole(self):
        return {
            "connected_tables": {"v2_connection_nodes": ["connection_node_id"]},
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "bottom_level": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Manhole bottom level.",
                "unit": "Meter above datum",
                "default": None,
            },
            "display_name": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Name field, no constraints.",
                "unit": None,
                "default": None,
            },
            "code": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Name field, no constraints.",
                "unit": None,
                "default": None,
            },
            "calculation_type": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [0, 1, 2],
                "description": """ 
                                Manhole calculation type for 1D-2D connection.
                                0: embedded
                                1: isolated
                                2: connected""",
                "unit": "L/day per inhabitant",
                "default": None,
            },
            "connection_node_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "ID of connection ndoe on which manhole is placed",
                "unit": None,
                "default": None,
            },
            "drain_level": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "Manhole drain level (**for connected manholes). If there is a connected manhole without drain level, 3Di will take the top of the pipe from the connection pipes as drain level.",
                "unit": "Meter above datum",
                "default": None,
            },
            "length": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "Manhole length. This value is for administrative purposes only and has no effect on the storage area of the connection node. **Mandatory when shape = 02",
                "unit": "Meter",
                "default": None,
            },
            "zoom_category": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": [0, 1, 2, 3, 4, 5],
                "description": "Name field, no constraints.",
                "unit": None,
                "default": None,
            },
            "manhole_indicator": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [0, 1, 2],
                "description": """  0: inspection (inspectieput) 1: outlet 2: pump""",
                "unit": None,
                "default": None,
            },
            "shape": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": """Manhole shape. This value is for administrative purposes only and has no effect on the storage area of the connection node. To add storage to a connection node, adjust the 'storage_area' in the v2_connection_nodes table.
                                00: square
                                01: round
                                02: rectangle """,
                "unit": None,
                "default": None,
            },
            "surface_level": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "Manhole surface level.",
                "unit": "Meter above datum",
                "default": None,
            },
            "sediment_level": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "Manhole surface level.",
                "unit": "Meter above datum",
                "default": None,
            },
            "width": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "Manhole width or diameter. This value is for administrative purposes only and has no effect on the storage area of the connection node.",
                "unit": "Meter",
                "default": None,
            },
        }

    @property
    def v2_numerical_settings(self):
        return {
            "connected_tables": {},
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "cfl_strictness_factor_1d": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "Strictness of CFL for 1D",
                "unit": None,
                "default": 1,
            },
            "cfl_strictness_factor_2d": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "Strictness of CFL for 2D",
                "unit": None,
                "default": 1,
            },
            "convergence_cg": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": """For numerical computation several tresholds are needed in the code, 
                    to avoid deficiencies due to a limited numerical accuracy. Generally this is to keep the behaviour consistent.""",
                "unit": None,
                "default": 0.000000001,
            },
            "convergence_eps": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": """Minimal residual for convergence of newton iteration.""",
                "unit": None,
                "default": 0.0001,
            },
            "flow_direction_threshold": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": """For numerical computation several tresholds are needed in the code, 
                    to avoid deficiencies due to a limited numerical accuracy. Generally this is to keep the behaviour consistent.""",
                "unit": "Meter per second",
                "default": 0.000001,
            },
            "frict_shallow_water_correction": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": """In case the friction assumptions based on the dominant friction balance gives a structurally underestimation of the friction, one can switch this setting on. 
                                0 is off, 1 is maximum between averaged friction and divided channel based friction, 2 is always linearized, 3 linearizes the depth based on a weighed averaged.  In this case the maximum depth of a thin layer needs to be defined. Do not use in combination with interflow.""",
                "unit": None,
                "default": 0,
            },
            "general_numerical_threshold": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": """For numerical computation several tresholds are needed in the code, 
                    to avoid deficiencies due to a limited numerical accuracy. Generally this is to keep the behaviour consistent.""",
                "unit": None,
                "default": 0.00000001,
            },
            "integration_method": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": """Time integration method: 0=Euler implicit.""",
                "unit": 0,
                "default": 0.000000001,
            },
            "limiter_grad_1d": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": """The limiter on the water level gradient allows the model to deal with unrealistically steep gradients.""",
                "unit": None,
                "default": 1,
            },
            "limiter_grad_2d": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": """The limiter on the water level gradient allows the model to deal with unrealistically steep gradients. When field is left empty, it is switched on!!!""",
                "unit": None,
                "default": 0,
            },
            "limiter_slope_crossectional_area_2d": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": """This limiter starts working in case the depth based on the downstream water level is zero and may be useful in sloping areas.
                                  0 is off, and 1 is a limiter which ends in a a higher order scheme, but is sensitive too instabilities, 2, treats the cross-sections as an upwind method volume/surface area under the assumption that the flow acts like a thin layer,  3 makes a combination of the traditional method in combination with the  thin layer apporach. In this case the maximum depth of a thin layer needs to be defined.
                                  Do not use in combination with interflow""",
                "unit": None,
                "default": 0,
            },
            "limiter_slope_friction_2d": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": """This limiter starts working in case the depth based on the downstream water level is zero and may be useful in sloping areas.  0 is off, and 1 is on. 
                            This limiter is obliged in combination with limiter_slope_crossectional_area_2d>0.. l Do not use in combination with interflow""",
                "unit": None,
                "default": 0,
            },
            "max_degree": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [700, 7, 5, 7, 70],
                "description": """Setting for matrix solver. Values below are adviced for different model types
                        700 for 1D flow
                        7 for 1D and 2D flow
                        5 for surface 2D flow only
                        7 for surface and groundwater flow
                        70 for 1D, 2D surface and groundwater flow or higher. Play around with this value in case of groundwater, can speed up your model significantly""",
                "unit": None,
                "default": 0,
            },
            "max_nonlin_iterations": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": """Maximum number of nonlinear iterations in single time step.""",
                "unit": None,
                "default": 20,
            },
            "minimum_friction_velocity": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": """For numerical computation several tresholds are needed in the code, to avoid deficiencies due to a limited numerical accuracy. Generally this is to keep the behaviour consistent..""",
                "unit": "m/s",
                "default": 0.05,
            },
            "minimum_surface_area": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": """Maximum number of nonlinear iterations in single time step.""",
                "unit": "m2",
                "default": 20,
            },
            "precon_cg": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": [0, 1],
                "description": """Use preconditioner for matrix solver. Increases simulation speed in most cases, Set to 0 or 1 (default).""",
                "unit": None,
                "default": 1,
            },
            "preissmann_slot": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": """A conceptual vertical and narrow slot providing a conceptual free surface condition for the flow when the water level is above the top of a closed conduit. Often used to guarentee stability, in 3Di unnessary unless used for presurized pipe flow. Works only for circular profiles.""",
                "unit": "m2",
                "default": 0,
            },
            "pump_implicit_ratio": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": [0, 1],
                "description": """Determines whether pump discharge is always maximum capacity (0) or discharge is limited to available inflow (1). The latter ensures a smooth discharge. Value between 0 and 1.""",
                "unit": None,
                "default": 1,
            },
            "thin_water_layer_definition": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": """** mandatory when using friction shallow water correction option 3 or limiter_slope_crossectional_area_2d on option 3""",
                "unit": "m",
                "default": 0.05,
            },
            "use_of_cg": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": """Number of iteration of conjugate gradient method, before swtiching to another method.""",
                "unit": "m",
                "default": 20,
            },
            "use_of_nested_newton": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": """1 for 1D calculation with closed profiles to handle non-linearity in volume-waterlevel relation. When using 0 nested newton is switch off by default but will be used when calculations become non-linear. For sewerage systems 1 is advised.""",
                "unit": "m",
                "default": 0,
            },
        }

    @property
    def v2_obstacle(self):
        return {
            "connected_tables": {},
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "code": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Name field, no constraints",
                "unit": None,
                "default": False,
            },
            "crest_level": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Crest level of obstacle segment.",
                "unit": "Meter above datum",
                "default": None,
            },
        }

    @property
    def v2_orifice(self):
        return {
            "connected_tables": {
                "v2_cross_section_definition": ["cross_section_definition_id"],
                "v2_connection_nodes": [
                    "connection_node_start_id",
                    "connection_node_end_id",
                ],
            },
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "display_name": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Name field, no constraints.",
                "unit": None,
                "default": None,
            },
            "code": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Name field, no constraints.",
                "unit": None,
                "default": None,
            },
            "zoom_category": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": [0, 1, 2, 3, 4, 5],
                "description": "Name field, no constraints.",
                "unit": None,
                "default": None,
            },
            "connection_node_start_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Start node for orifice",
                "unit": None,
                "default": None,
            },
            "connection_node_end_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "End node for orifice",
                "unit": None,
                "default": None,
            },
            "crest_level": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Crest or bottom level",
                "unit": "Meter above datum",
                "default": None,
            },
            "crest_type": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [3, 4],
                "description": """Type of weir formulation.
                                  3: broad crested
                                  4: short crested""",
                "unit": None,
                "default": 4,
            },
            "cross_section_definition_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "ID of cross section definiton in v2_cross_section_definition",
                "unit": None,
                "default": None,
            },
            "discharge_coefficient_negative": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Discharge coefficient for negative flow (from end to start node). Can be set to 0 when closed.",
                "unit": None,
                "default": 1,
            },
            "discharge_coefficient_positive": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Discharge coefficient for postive flow (from start to end node). Can be set to 0 when closed.",
                "unit": None,
                "default": 1,
            },
            "friction_type": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [1, 2],
                "description": """Friction Type.
                                1: Chezy (not yet implemented)
                                 2: Manning""",
                "unit": None,
                "default": 2,
            },
            "friction_value": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Friction or roughness value for profile",
                "unit": "1:[m1/2/s], 2:[s/m1/3]",
                "default": None,
            },
            "sewerage": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "For internal book keeping. Can be used for statistics in QGIS plugin.",
                "unit": None,
                "default": None,
            },
            # TODO
            # Has been removed after klondike
            # "max_capacity": {
            #     "type": ogr.OFTReal,
            #     "optional": True,
            #     "options": None,
            #     "description": "Not Needed",
            #     "unit": None,
            #     "default": None,
            # },
        }

    @property
    def v2_pipe(self):
        return {
            "connected_tables": {
                "v2_cross_section_definition": ["cross_section_definition_id"],
                "v2_connection_nodes": [
                    "connection_node_end_id",
                    "connection_node_start_id",
                ],
            },
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "display_name": {
                "type": ogr.OFTString,
                "optional": True,
                "options": None,
                "description": "Name field, no constraints.",
                "unit": None,
                "default": None,
            },
            "code": {
                "type": ogr.OFTString,
                "optional": True,
                "options": None,
                "description": "Name field, no constraints.",
                "unit": None,
                "default": None,
            },
            "zoom_category": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": [0, 1, 2, 3, 4, 5],
                "description": "Name field, no constraints.",
                "unit": None,
                "default": None,
            },
            "connection_node_start_id": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "Start node for pipe",
                "unit": None,
                "default": None,
            },
            "connection_node_end_id": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "End node for pipe",
                "unit": None,
                "default": None,
            },
            "calculation_type": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [0, 1, 2, 3, 4],
                "description": """Calculation type for pipe. When start en end connection nodes are manholes only used for calculation points half-way pipe.
                                    0 = embedded
                                    1 = isolated
                                    2 = connected
                                    3 = broad crest
                                    4 = short crest""",
                "unit": None,
                "default": 1,
            },
            "dist_calc_points": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": [3, 4],
                "description": """Type of weir formulation.
                                  3: broad crested
                                  4: short crested""",
                "unit": None,
                "default": 4,
            },
            "cross_section_definition_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "ID of cross section definiton in v2_cross_section_definition",
                "unit": None,
                "default": None,
            },
            "invert_level_start_point": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Invert level at pipe startpoint. ",
                "unit": None,
                "default": None,
            },
            "invert_level_end_point": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Invert level at pipe endpoint. ",
                "unit": None,
                "default": None,
            },
            "friction_type": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [1, 2],
                "description": """Friction Type.
                                1: Chezy (not yet implemented)
                                 2: Manning""",
                "unit": None,
                "default": 2,
            },
            "friction_value": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Friction or roughness value for profile",
                "unit": "1:[m1/2/s], 2:[s/m1/3]",
                "default": None,
            },
            "material": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": [0, 1, 2, 3, 4, 5, 6, 7, 8],
                "description": """Material of pipe, used for internal bookkeeping only.
                                  0: concrete
                                 1: pvc
                                 2: gres
                                 3: cast iron
                                 4: brickwork
                                 5: HPE
                                 6: HDPE
                                 7: plate iron
                                 8: steel""",
                "unit": None,
                "default": None,
            },
            "original_length": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "For internal book keeping.",
                "unit": "Meter",
                "default": None,
            },
            "profile_num": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "For internal book keeping.",
                "unit": None,
                "default": None,
            },
            "sewerage_type": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [0, 1, 2, 3, 4, 5, 6, 7],
                "description": """ 
                            Pipe type. 3Di requires the sewerage_type to be one of the following:
                            0: gemengd - mixed
                             1: rwa - rain water
                             2: dwa - dry wheather flow
                             3: transport
                             4: overstort - spillway
                             5: zinker
                             6: berging - storage
                             7: bergbezinkbak - storage tank
                             """,
                "unit": None,
                "default": None,
            },
        }

    @property
    def v2_pumpstation(self):
        return {
            "connected_tables": {
                "v2_connection_nodes": [
                    "connection_node_start_id",
                    # "connection_node_end_id",
                ]
            },
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "display_name": {
                "type": ogr.OFTString,
                "optional": True,
                "options": None,
                "description": "Name field, no constraints.",
                "unit": None,
                "default": None,
            },
            "code": {
                "type": ogr.OFTString,
                "optional": True,
                "options": None,
                "description": "Name field, no constraints.",
                "unit": None,
                "default": None,
            },
            "capacity": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Pump capacity.",
                "unit": "Liters per second",
                "default": None,
            },
            "zoom_category": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": [0, 1, 2, 3, 4, 5],
                "description": "Name field, no constraints.",
                "unit": None,
                "default": None,
            },
            "connection_node_start_id": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "Start node for pumpstation",
                "unit": None,
                "default": None,
            },
            "connection_node_end_id": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "End node for pumpstation",
                "unit": None,
                "default": None,
            },
            "classification": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "TODO",
                "unit": None,
                "default": None,
            },
            "lower_stop_level": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Level at pump start or end node at which pump stops pumping.",
                "unit": "Meter above datum",
                "default": None,
            },
            "sewerage": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "For internal book keeping. Specifies if the pumpstation is in sewerage",
                "unit": None,
                "default": None,
            },
            "start_level": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Level at pump start or end node from from which it starts pumping.",
                "unit": "Meter above datum",
                "default": None,
            },
            "type": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [1, 2],
                "description": """
                            Type that determines pump function.
                            1 Pump behaviour is based on water levels on the suction-side of the pump
                            2 Pump behaviour is based on water levels on the delivery-side of the pump  """,
                "unit": None,
                "default": None,
            },
            "upper_stop_level": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "Level at pump start or end node at which pump stops pumping.Must be above start level.",
                "unit": "Meter above datum",
                "default": None,
            },
        }

    @property
    def v2_simple_infiltration(self):
        return {
            "connected_tables": {},
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "display_name": {
                "type": ogr.OFTString,
                "optional": True,
                "options": None,
                "description": "Name field, no constraints.",
                "unit": None,
                "default": None,
            },
            "infiltration_rate": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "global infiltration raste",
                "unit": "mm/day",
                "default": 0,
            },
            "infiltration_rate_file": {
                "type": ogr.OFTString,
                "optional": True,
                "options": None,
                "description": "Relative path to infiltration file (.tif).",
                "unit": "mm/day",
                "default": None,
            },
            "infiltration_surface_option": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": [0, 1, 2],
                "description": """Option that sets how the infiltration works in calculation cells.
                                0: rain (whole surface when raining, only wet pixels when dry)
                                1: whole surface (always whole surface)
                                2: only wet surface (always only wet pixels)
                                in case not defined then option 0 is used
                                """,
                "unit": None,
                "default": 0,
            },
            "max_infiltration_capacity_file": {
                "type": ogr.OFTString,
                "optional": True,
                "options": [1, 2],
                "description": """Relative path to max infiltration file (.tif). Maximum infiltration uses the sum of pixel values per calculation cell.""",
                "unit": "Meter",
                "default": None,
            },
        }

    @property
    def v2_surface(self):
        return {
            "connected_tables": {"v2_surface_parameters": ["surface_parameters_id"]},
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "area": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "area of surface",
                "unit": None,
                "default": None,
            },
            "display_name": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Name field, no constraints.",
                "unit": None,
                "default": None,
            },
            "code": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Name field, no constraints.",
                "unit": None,
                "default": None,
            },
            "dry_weather_flow": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "Dry weather flow per inhabitant",
                "unit": "L/day per inhabitant",
                "default": None,
            },
            "nr_of_inhabitants": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "Number of inhabitant used for dry wheather flow.",
                "unit": "Nr of inhabitants",
                "default": None,
            },
            "function": {
                "type": ogr.OFTString,
                "optional": True,
                "options": None,
                "description": """For your own administration.""",
                "unit": None,
                "default": None,
            },
            "surface_parameters_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": """Reference to v2_surface_parameters. The id filled in here must be present in this (v2_surface_parameters) table """,
                "unit": None,
                "default": None,
            },
            "zoom_category": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": [0, 1, 2, 3, 4, 5],
                "description": "Name field, no constraints.",
                "unit": None,
                "default": None,
            },
        }

    @property
    def v2_surface_map(self):
        return {
            "connected_tables": {
                "v2_surface": ["surface_id"],
                "v2_connection_nodes": ["connection_node_id"],
            },
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "connection_node_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "ID of connection_node. Connecting the surface area from v2_surface to an connection_node",
                "unit": None,
                "default": None,
            },
            "percentage": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "percentage of area to the connection_node",
                "unit": "%",
                "default": None,
            },
            "surface_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "ID of surface feature",
                "unit": None,
                "default": None,
            },
            "surface_type": {
                "type": ogr.OFTString,
                "optional": False,
                "options": ["v2_surface", "v2_impervious_surface"],
                "description": "choice to use the 'v2_surface' or 'v2_impervious_surface'",
                "unit": None,
                "default": None,
            },
        }

    @property
    def v2_surface_parameters(self):
        return {
            "connected_tables": {},
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "infiltration": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [0, 1],
                "description": "0 or 1",
                "unit": None,
                "default": None,
            },
            "infiltration_decay_constant": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "time factor decay infiltration capacity of the surface",
                "unit": "/hour",
                "default": None,
            },
            "infiltration_recovery_constant": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "time factor recovery infiltration capacity of the surface",
                "unit": "/hour",
                "default": None,
            },
            "max_infiltration_capacity": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Maximum infiltration capacity of the surface",
                "unit": "mm/h",
                "default": None,
            },
            "min_infiltration_capacity": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Minimum infiltration capacity of the surface",
                "unit": "mm/h",
                "default": None,
            },
            "outflow_delay": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "delay of outflow",
                "unit": "/min",
                "default": None,
            },
            "surface_layer_thickness": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "mm storage on the surface",
                "unit": "mm",
                "default": None,
            },
        }

    @property
    def v2_weir(self):
        return {
            "connected_tables": {
                "v2_cross_section_definition": ["cross_section_definition_id"],
                "v2_connection_nodes": [
                    "connection_node_start_id",
                    "connection_node_end_id",
                ],
            },
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "display_name": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Name field, no constraints.",
                "unit": None,
                "default": None,
            },
            "code": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Name field, no constraints.",
                "unit": None,
                "default": None,
            },
            "zoom_category": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": [0, 1, 2, 3, 4, 5],
                "description": "Name field, no constraints.",
                "unit": None,
                "default": None,
            },
            "connection_node_start_id": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "Start node for weir",
                "unit": None,
                "default": None,
            },
            "connection_node_end_id": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "End node for weir",
                "unit": None,
                "default": None,
            },
            "crest_level": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Crest or bottom level",
                "unit": "Meter above datum",
                "default": None,
            },
            "crest_type": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [3, 4],
                "description": """Type of weir formulation.
                                  3: broad crested
                                  4: short crested""",
                "unit": None,
                "default": 4,
            },
            "cross_section_definition_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "ID of cross section definiton in v2_cross_section_definition",
                "unit": None,
                "default": None,
            },
            "discharge_coefficient_negative": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Discharge coefficient for negative flow (from end to start node). Can be set to 0 when closed.",
                "unit": None,
                "default": 1,
            },
            "discharge_coefficient_positive": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Discharge coefficient for postive flow (from start to end node). Can be set to 0 when closed.",
                "unit": None,
                "default": 1,
            },
            "friction_type": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [1, 2],
                "description": """Friction Type.
                                1: Chezy (not yet implemented)
                                 2: Manning""",
                "unit": None,
                "default": 2,
            },
            "friction_value": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Friction or roughness value for profile",
                "unit": "1:[m1/2/s], 2:[s/m1/3]",
                "default": None,
            },
            "sewerage": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "For internal book keeping. Can be used for statistics in QGIS plugin.",
                "unit": None,
                "default": None,
            },
            "external": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "For internal book keeping. Can be used for statistics in QGIS plugin.",
                "unit": None,
                "default": None,
            },
        }

    @property
    def v2_calculation_point(self):
        return {
            "id": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "content_type_id": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "Content type",
                "unit": None,
                "default": None,
            },
            "user_ref": {
                "type": ogr.OFTString,
                "optional": True,
                "options": None,
                "description": "Reference",
                "unit": None,
                "default": None,
            },
            "calc_type": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "Calculation type",
                "unit": None,
                "default": None,
            },
            "the_geom": {
                "type": ogr.wkbPoint,
                "optional": False,
                "options": None,
                "description": "Geometry",
                "unit": "Depend on projection",
                "default": None,
            },
        }

    @property
    def v2_connected_pnt(self):
        return {
            "id": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "exchange_level": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "exchange level for leveees",
                "unit": None,
                "default": None,
            },
            "calculation_pnt_id": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "Caculation point id",
                "unit": None,
                "default": None,
            },
            "levee_id": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "id of levee",
                "unit": None,
                "default": None,
            },
            "the_geom": {
                "type": ogr.wkbPoint,
                "optional": False,
                "options": None,
                "description": "Geometry",
                "unit": "Depend on projection",
                "default": None,
            },
        }

    @property
    def v2_control(self):
        return {
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "control_type": {
                "type": ogr.OFTString,
                "optional": False,
                "options": ["table", "memory", "timed"],
                "description": "Type of control, options are amongst others: table, memory",
                "unit": None,
                "default": None,
            },
            "control_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "id in the v2_control_table (in case of table control) or v2_control_memory (in case of memory control)",
                "unit": None,
                "default": None,
            },
            "control_group_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "id of the v2_control_group this control is part of",
                "unit": None,
                "default": None,
            },
            "measure_group_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "id of the v2_measure_group",
                "unit": None,
                "default": None,
            },
            "start": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "Start time of the control in seconds since beginning of the simulation. Can be used to link multiple control tables to one control structure.",
                "unit": "seconds",
                "default": None,
            },
            "end": {
                "type": ogr.OFTReal,
                "optional": True,
                "options": None,
                "description": "End time of the control in seconds since beginning of the simulation",
                "unit": None,
                "default": None,
            },
            "measure_frequency": {
                "type": ogr.OFTInteger,
                "optional": True,
                "options": None,
                "description": "NOT YET IMPLEMENTED",
                "unit": None,
                "default": None,
            },
        }

    @property
    def v2_control_group(self):
        return {
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "name": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Name",
                "unit": None,
                "default": None,
            },
            "description": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Description",
                "unit": None,
                "default": None,
            },
        }

    @property
    def v2_control_measure_group(self):
        return {
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            }
        }

    @property
    def v2_control_measure_map(self):
        return {
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "measure_group_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "ID of the v2_measure_group this measure station is part of.",
                "unit": None,
                "default": None,
            },
            "object_type": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Type of object to measure at, for example:'v2_connection_nodes'",
                "unit": None,
                "default": None,
            },
            "object_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "id of the object (of type defined in object_type)",
                "unit": None,
                "default": None,
            },
            "weight": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "weight of measuring station in group, use 1.0 for groups with single measuring station. Combined weight should be 1.0.",
                "unit": None,
                "default": None,
            },
        }

    @property
    def v2_control_table(self):
        return {
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "action_table": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": """Semicolumn seperated table with action values, use # for newline. For Example: -1.7;-1.4#-1.6;-1.3#-1.5;-1.2
                                  When controlling set_discharge_coefficients you need to supply 2 values. One for the positive discharge coefficient and one for the negative discharge coefficient. Example: -1.7;0 0#-1.6;0.5 0.8#-1.5;1 1
                               """,
                "unit": None,
                "default": None,
            },
            "action_type": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Type of action; For instance: 'set_crest_level' or 'set_discharge_coefficients'",
                "unit": None,
                "default": None,
            },
            "measure_variable": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Measure variable in action table. For instance: 'waterlevel'",
                "unit": None,
                "default": None,
            },
            "measure_operator": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Operator for direction the action table is read. '<' or '>'",
                "unit": None,
                "default": None,
            },
            "target_type": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Structure type the control is applied to. For instance: 'v2_weir' or 'v2_culvert' or 'v2_orifice'",
                "unit": None,
                "default": None,
            },
            "target_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Id of structure the control is applied to.",
                "unit": None,
                "default": None,
            },
        }

    @property
    def v2_control_memory(self):
        return {
            "id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Unique identifier",
                "unit": None,
                "default": None,
            },
            "action_value": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": """Value that the measure_variable is set to when memory control becomes active""",
                "unit": None,
                "default": None,
            },
            "action_type": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Type of action; For instance: set_pump_capacity",
                "unit": None,
                "default": None,
            },
            "is_active": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [0, 1],
                "description": """0: control is inactive when initializing the model
                                  1: control is active when initializing the model""",
                "unit": None,
                "default": None,
            },
            "is_inverse": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": [0, 1],
                "description": """
                                0: normal functioning of the control
                                1: inverting the lower and upper threshold
                            """,
                "unit": None,
                "default": None,
            },
            "lower_threshold": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Lower threshold of measure_variable. Control becomes inactive when value drops below this value (unless is_inverse = 1)",
                "unit": None,
                "default": None,
            },
            "measure_variable": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Measure variable in action table. For instance: waterlevel",
                "unit": None,
                "default": None,
            },
            "target_type": {
                "type": ogr.OFTString,
                "optional": False,
                "options": None,
                "description": "Structure type the control is applied to. For instance: 'v2_weir' or 'v2_culvert' or 'v2_orifice'",
                "unit": None,
                "default": None,
            },
            "target_id": {
                "type": ogr.OFTInteger,
                "optional": False,
                "options": None,
                "description": "Id of structure the control is applied to.",
                "unit": None,
                "default": None,
            },
            "upper_threshold": {
                "type": ogr.OFTReal,
                "optional": False,
                "options": None,
                "description": "Upper threshold of measure_variable. Control becomes active when value rises above this value (unless is_inverse = 1)",
                "unit": None,
                "default": None,
            },
        }

    @property
    def v2_control_delta(self):
        return {}  # TODO

    @property
    def v2_control_pid(self):
        return {}  # TODO

    @property
    def v2_control_timed(self):
        return {}  # TODO

    @property
    def v2_windshielding(self):
        return {}  # TODO

    @property
    def v2_dem_average_area(self):
        return {}  # TODO

    @property
    def v2_floodfill(self):
        return {}  # TODO

    @property
    def v2_pumped_drainage_area(self):
        return {}  # TODO

    @property
    def v2_initial_waterlevel(self):
        return {}  # TODO

    @property
    def v2_outlet(self):
        return {}  # TODO


TRANSLATE = {
    "v2_global_settings": "global_settings",
    "v2_connection_nodes": "nodes",
    "v2_manhole": "manholes",
    "v2_pipe": "pipes",
    "v2_simple_infiltration": "simple_infiltration",
    "v2_numerical_settings": "numerical_settings",
    "v2_aggregation_settings": "aggregation_settings",
    "v2_obstacle": "obstacles",
    "v2_channel": "channels",
    "v2_cross_section_definition": "cross_section_definitions",
    "v2_cross_section_location": "cross_section_locations",
    "v2_grid_refinement_area": "grid_refinement_areas",
    "v2_grid_refinement": "grid_refinements",
    "v2_1d_boundary_conditions": "boundary_conditions_1d",
    "v2_1d_lateral": "laterals_1d",
    "v2_pumpstation": "pumpstations",
    "v2_levee": "levees",
    "v2_weir": "weirs",
    "v2_orifice": "orifices",
    "v2_culvert": "culverts",
    "v2_calculation_point": "calculation_points",
    "v2_connected_pnt": "connection_points",
    "v2_control": "control",
    "v2_control_group": "control_group",
    "v2_control_measure_group": "control_measure_group",
    "v2_control_measure_map": "control_measure_map",
    "v2_control_table": "control_table",
    "v2_control_memory": "control_memory",
    "v2_impervious_surface": "impervious_surface",
    "v2_interflow": "interflow",
}



# -*- coding: utf-8 -*-
"""
Created on Mon Feb 15 17:02:45 2021

@author: chris.kerklaan



"""
# First-party imports
import logging
import pathlib

# Third-party imports
import numpy as np
from osgeo import gdal
from osgeo import ogr

# local imports
from .raster import Raster, loc_dir
from .vector import Vector
from threedi_edits.utils.project import Progress
from threedi_edits.utils.project import Classes, Functions

# structure
classes = Classes(__name__, local_only=True)
functions = Functions(__name__)


# GLOBALS
logger = logging.getLogger(__name__)
EPSG = 28992
DATA_TYPE = gdal.GDT_Float32
NODATA = -9999
NP_DATA_TYPE = "f4"
PLACEMENT = "stacked"
BLOCKSIZE = None

# DRIVERS
DRIVER_GDAL_GTIFF = gdal.GetDriverByName("GTiff")

# Write options
WRITE_OPTIONS = ["COMPRESS=DEFLATE", "NUM_THREADS=4", "ZLEVEL=9", "TILED=YES"]


class RasterGroup:
    """
    RasterGroup -- info
    RasterGroup is a class which makes handling multiple rasters easier.
    Multiple placement types of rastersgroups are available:
        - stacked (default)
        - adjacent

    Functions include:
            - Aligning all rasters
            - Finding the maximum extent
            - Clipping all rasters by  a vector file

            Rastergroup loops:
                for tiles in group.tiles:
                    tiles[0]# raster 0
                    tiles[1] # raster 1
                    etc.

            Inflated rastrgroup loops:
                inflated_tile_generator = group.inflated_tiles(buffer)
                for tiles, extent in zip(inflated_tile_generator, group.tile_extents):
                    # do something with tiles
                    tiles[0].array = tiles[0].array * tiles[1].array
                    clipped = tiles[0].clip_bbox(*extent)
                    clipped.write("something.tif")

    Params:
        rasters: list of Raster objects
        epsg: epsg format
        nodata: standard nodata for all rasters
        data_type: data_type for all rasters
        np_data_type: numpy variant of the data type

    """

    instance = "gis.rastergroup.RasterGroup"

    def __init__(
        self,
        rasters: list,
        placement=PLACEMENT,
        epsg=EPSG,
        nodata=NODATA,
        data_type=DATA_TYPE,
        np_data_type=NP_DATA_TYPE,
        blocksize=BLOCKSIZE,
    ):
        self.rasters = rasters
        self.names = [raster.name for raster in self.rasters]
        self.nodata = nodata
        self.epsg = epsg
        self.data_type = data_type
        self.np_data_type = np_data_type
        self.placement = placement
        self._blocksize = blocksize

        if len(self.rasters) > 0:
            self.blocksize = rasters[0].blocksize

        for name, raster in zip(self.names, self.rasters):
            setattr(self, name, raster)

    @classmethod
    def from_paths(
        cls,
        paths,
        epsg=EPSG,
        nodata=NODATA,
        data_type=DATA_TYPE,
        np_data_type=NP_DATA_TYPE,
    ):
        return cls([Raster(p) for p in paths])

    def __iter__(self):
        for raster in self.rasters:
            yield raster

    def __len__(self):
        return sum([len(raster) for raster in self])

    def __getitem__(self, name):
        if type(name) == "str":
            return self.rasters[self.names.index(name)]
        elif type(name) == int:
            return self.rasters[name]

    def __setitem__(self, name, value):
        """without alignemnt"""
        if name in self.names:
            self.rasters[self.names.index(name)] = value
        else:
            self.rasters.append(value)
            self.names.append(name)

        setattr(self, name, value)

    def __repr__(self):
        data = {"aligned": self.aligned}
        return f"({self.instance}) {self.names}: {data}"

    @classmethod
    def from_folder(cls, path):
        rasters = [r for r in pathlib.Path(path).rglob("*.tif")]
        return cls([Raster(str(r)) for r in rasters])

    @property
    def epsg(self):
        if not hasattr(self, "_epsg"):
            self._epsg = self[0].epsg
        return self._epsg

    @epsg.setter
    def epsg(self, value):
        self._epsg = value

    @property
    def count(self):
        return len(self.rasters)

    @property
    def aligned(self):
        self.check_alignment(False)
        if len(self._alignment["to_be_aligned"]) == 0:
            return True
        else:
            return False

    @property
    def aligned_nodata(self):
        self.check_alignment(True)
        if len(self._alignment["to_be_aligned"]) == 0:
            return True
        else:
            return False

    @property
    def tile_extents(self):
        return self.rasters[0].extent

    @property
    def extents(self):
        if not hasattr(self, "_extent"):
            _extent = Vector.from_scratch("extent", ogr.wkbPolygon, self.epsg)
            _extent.add_field("name", str)
            _extent.add_field("index", int)
            for i, raster in enumerate(self):
                _extent.add(
                    geometry=raster.extent_geometry,
                    items={"name": raster.name, "index": i},
                )
            self._extent = _extent
        return self._extent

    @property
    def stacked(self):
        return self.placement == "stacked"

    @stacked.setter
    def stacked(self, value):
        assert value in [True, False]
        if value == True:
            self.placement = "stacked"
        else:
            self.adjacent = True

    @property
    def adjacent(self):
        return self.placement == "adjacent"

    @adjacent.setter
    def adjacent(self, value):
        assert value in [True, False]
        if value == True:
            self.placement = "adjacent"
        else:
            self.stacked = True

    @property
    def blocksize(self):
        return self._blocksize

    @blocksize.setter
    def blocksize(self, blocksize):
        self._blocksize = blocksize
        for r in self:
            r.blocksize = blocksize
            r.optimize_blocksize = False

    def read(self, geometry):
        rasters = self.extents.spatial_filter(geometry)
        if len(rasters) == 0:
            return [np.nan]
        else:
            data = []
            for raster in rasters:
                data.append(self.rasters[raster["index"]].read(geometry))
            return data

    def tiles_inflated(self, buffer_size, original_extent):
        """
        Returns a generator based on the placement of the rastergroup.
        Returns multiple raster tiles if stacked.
        Returns a single raster tile if adjacent.
        """

        if self.stacked:
            if not self.aligned:
                raise ValueError("Not aligned, please align first")

            generators = [r.tiles_inflated(buffer_size) for r in self]

            for inflated_tiles in zip(*generators):
                yield inflated_tiles

        # TODO tiles cannot be inflated in this way.
        if self.adjacent:
            for raster in self:
                for tile in raster.tiles_inflated(buffer_size, original_extent):
                    yield tile

    def tiles(self):
        """
        Returns a generator based on the placement of the rastergroup.
        If stacked, we return mulitple raster tiles.
        If adjacent, we return a single raster tile.
        """

        if self.stacked:
            if not self.aligned:
                raise ValueError("Not aligned, please align first")

            for raster_tiles in zip(*self.rasters):
                yield raster_tiles

        if self.adjacent:
            for raster in self.rasters:
                for tile in raster:
                    yield tile

    def check_alignment(self, count_nodata=True):
        logger.info("Checking alignment of the rasters")
        self._alignment = check_alignment(self.rasters, count_nodata)
        return self._alignment

    def get_maximum_extent(self):
        if not hasattr(self, "maximum_extent"):
            self.maximum_extent = max_template(self.rasters)

    def align(self, index=0, use_maximum_extent=False, nodata=False):
        """
        This function aligns all rasters in the group
        params:
            index: This index is used for the raster on which all other
            rasters are aligned.
            use_maximum_extent: Calculates a maximum extent of all rasters
            and aligns all rasters to that maximum extent

        """
        logger.info("Aligning all rasters")

        if use_maximum_extent:
            self.get_maximum_extent()
            align_raster = self.maximum_extent
            rasters = self.rasters
        else:
            align_raster = self.rasters[index]
            rasters = [r for i, r in enumerate(self.rasters) if i != index]

        aligned_rasters = [r.align(align_raster, nodata_align=nodata) for r in rasters]

        if not use_maximum_extent:
            aligned_rasters.insert(index, align_raster)

        self.rasters = aligned_rasters

    def merge(self):
        """
        Merges the rasters based on the maximum extent
        loops over every tile and adds data on top of the nodata of the other
        """
        logger.info("start merging")

        # they all must be align, have the same extent an the same nodata value
        self.get_maximum_extent()
        self.align(use_maximum_extent=True)
        for raster in self.rasters:
            raster.optimize_blocksize = False

        pbar = Progress(len(raster), "Merging")

        target = self.rasters[0].copy()
        for tiles in zip(*tuple(self.rasters)):
            tile_array = tiles[0].array

            for tile in tiles[1:]:
                mask = np.isnan(tile_array)
                tile_array[mask] = tile.array[mask]

            pbar.update()
            target.array = tile_array, *tile.location

        return target

    def vrt(self):
        """
        Creates one raster out of all adjacent rasters.
        Returns a vrt of all rasters.


        """
        raster_list = [r.ds for r in self.rasters]
        path = loc_dir(mem=True).replace(".tif", ".vrt")
        vrt_options = gdal.BuildVRTOptions(resampleAlg="cubic")
        ds = gdal.BuildVRT(path, raster_list, options=vrt_options)
        return Raster(ds)

    def insert(self, raster: Raster):
        logger.info(f"Inserting a {raster.name} into the group")
        align_raster = self.rasters[0]
        raster.align(align_raster, nodata_align=False)
        self.rasters.append(raster)
        self.names.append(raster.name)

    def clip(self, vector):
        """clips all available rasters to the vector file"""
        for raster in self:
            logger.debug(f"Clipping {raster.name}")
            clipped = raster.clip(vector)
            self[raster.name] = clipped

    def write(self, filename):
        """writes it to one dataset, with different layers"""

        ds = DRIVER_GDAL_GTIFF.Create(filename, options=WRITE_OPTIONS)
        ds.GetRasterBand(1).SetNoDataValue(self.nodata_value)
        ds.SetProjection(self.spatial_reference.wkt)
        ds.SetGeoTransform(self.geotransform)
        ds.SetMetadata(self.metadata)


def check_alignment(raster_list, count_nodata=True):
    """checks data/nodata alignemnt and extent alignment"""
    dem = raster_list[0]

    output = {
        "extent": {},
        "counts": {},
        "location": {},
        "errors": [],
        "to_be_aligned": [],
    }
    for raster in raster_list:
        # extent
        if (raster.columns, raster.rows) != (dem.columns, dem.rows):
            msg = f"{raster.name} has unqual columns and rows "
            logger.debug(msg)
            output["errors"].append(("extent", msg))
            if not raster.name in output["to_be_aligned"]:
                output["to_be_aligned"].append(raster.name)

        output["extent"][raster.name] = {
            "rows": raster.rows,
            "columns": raster.columns,
        }
        if count_nodata:
            # data/nodata
            output["counts"][raster.name] = count_data_nodata(raster)

        # location
        if raster.geotransform != dem.geotransform:
            msg = f"{raster.name} has not a similar geotransform as the first raster"
            logger.debug(msg)
            output["errors"].append(("location", msg))
            if not raster.name in output["to_be_aligned"]:
                output["to_be_aligned"].append(raster.name)

        output["location"][raster.name] = raster.geotransform

    for key, values in output["counts"].items():
        if values != output["counts"][dem.name]:
            msg = f"{key} pixel data/nodata count not equal"
            logger.debug(msg)
            output["errors"].append(("counts", msg))
            if not key in output["to_be_aligned"]:
                output["to_be_aligned"].append(key)

    if len(output["errors"]) == 0:
        logger.debug("RasterGroup - Check alignment found no problems")

    return output


def count_data_nodata(raster):
    """input array has np.nan as nodata"""
    count_data = 0
    count_nodata = 0
    raster.optimize_blocksize = False
    for data in raster:
        arr = data.array
        total_size = arr.size
        add_cnt_nodata = np.count_nonzero(np.isnan(arr))
        add_cnt_data = total_size - add_cnt_nodata
        count_nodata += add_cnt_nodata
        count_data += add_cnt_data
        del arr
    return count_data, count_nodata


def max_template(rasters=[Raster, Raster], resolution=None):
    """Takes a list of Raster objects and returns the largest template
    Note that the resolution of the first raster is taken, if not overwritten
    by the resolution param.
    """

    x_min = 1e31
    x_max = -1e31
    y_min = 1e31
    y_max = -1e31

    if resolution == None:
        resolution = rasters[0].resolution["width"]

    for raster in rasters:

        gtt = raster.geotransform
        # xmin
        if gtt[0] < x_min:
            x_min = gtt[0]

        # xmax
        if gtt[0] + (raster.rows * gtt[1]) > x_max:
            x_max = gtt[0] + (raster.rows * gtt[1])

        # ymax
        if gtt[3] > y_max:
            y_max = gtt[3]

        # ymin
        if gtt[3] + (raster.columns * gtt[5]) < y_min:
            y_min = gtt[3] + (raster.columns * gtt[5])

    rows = (y_max - y_min) / resolution
    columns = (x_max - x_min) / resolution
    template = Raster.from_array(np.zeros((int(columns), int(rows))))
    template.geotransform = (x_min, resolution, 0.0, y_max, 0.0, -resolution)
    template.nodata_value = -9999
    template.spatial_reference = raster.epsg

    return template



# -*- coding: utf-8 -*-
"""
Created on Tue Feb 16 14:52:59 2021
GDAL utils from raster-tools.

"""
from osgeo import ogr
from osgeo import gdal
from osgeo import gdal_array


def create(
    array,
    band=False,
    geotransform=None,
    spatial_reference=None,
    nodata_value=None,
):
    """
    Create and return a gdal dataset.

    :param array: A numpy array.
    :param geo_transform: 6-tuple of floats
    :param projection: wkt projection string
    :param no_data_value: integer or float

    This is the fastest way to get a gdal dataset from a numpy array, but
    keep a reference to the array around, or a segfault will occur. Also,
    don't forget to call FlushCache() on the dataset after any operation
    that affects the array.
    """
    # prepare dataset name pointing to array
    datapointer = array.ctypes.data
    if len(array.shape) == 2:
        lines, pixels = array.shape
        bands = 1
    else:
        bands, lines, pixels = array.shape
    datatypecode = gdal_array.NumericTypeCodeToGDALTypeCode(array.dtype.type)
    datatype = gdal.GetDataTypeName(datatypecode)
    if len(array.strides) == 2:
        lineoffset, pixeloffset = array.strides
        bandoffset = 0
    else:
        bandoffset, lineoffset, pixeloffset = array.strides

    dataset_name_template = (
        "MEM:::"
        "DATAPOINTER={datapointer},"
        "PIXELS={pixels},"
        "LINES={lines},"
        "BANDS={bands},"
        "DATATYPE={datatype},"
        "PIXELOFFSET={pixeloffset},"
        "LINEOFFSET={lineoffset},"
        "BANDOFFSET={bandoffset}"
    )
    dataset_name = dataset_name_template.format(
        datapointer=datapointer,
        pixels=pixels,
        lines=lines,
        bands=bands,
        datatype=datatype,
        pixeloffset=pixeloffset,
        lineoffset=lineoffset,
        bandoffset=bandoffset,
    )

    # access the array memory as gdal dataset
    dataset = gdal.Open(dataset_name, gdal.GA_Update)

    # set additional properties from kwargs
    if geotransform is not None:
        dataset.SetGeoTransform(geotransform)
    if spatial_reference is not None:
        dataset.SetProjection(spatial_reference)
    if nodata_value is not None:
        dataset.GetRasterBand(1).SetNoDataValue(nodata_value)
        # for i in range(len(array)):
        # dataset.GetRasterBand(i + 1).SetNoDataValue(no_data_value)

    if band:
        return dataset, dataset.GetRasterBand(1)
    else:
        return dataset


class Dataset(object):
    """
    Usage:
        >>> with Dataset(array) as dataset:
        ...     # do gdal things.
    """

    def __init__(self, array, **kwargs):
        self.array = array
        self.dataset = create(array, False, **kwargs)

    def __enter__(self):
        return self.dataset

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        self.dataset.FlushCache()


class Layer:
    """
    Usage:
        >>> with Layer(geometry) as layer:
        ...     # do ogr things.
    """

    def __init__(self, geometry):
        driver = ogr.GetDriverByName(str("Memory"))
        self.data_source = driver.CreateDataSource("")
        sr = geometry.GetSpatialReference()
        self.layer = self.data_source.CreateLayer(str(""), sr)
        layer_defn = self.layer.GetLayerDefn()
        feature = ogr.Feature(layer_defn)
        feature.SetGeometry(geometry)
        self.layer.CreateFeature(feature)

    def __enter__(self):
        return self.layer

    def __exit__(self, exc_type, exc_value, traceback):
        pass


DRIVERS = {}
for i in range(ogr.GetDriverCount()):
    drv = ogr.GetDriver(i)
    if (
        drv.GetMetadataItem("DCAP_VECTOR") == "YES"
        and drv.GetMetadataItem("DCAP_CREATE") == "YES"
    ):
        driver = drv.GetMetadataItem("DMD_EXTENSION")
        if driver:
            DRIVERS["." + driver] = drv.GetName()



# # -*- coding: utf-8 -*-
# """
# Created on Fri Feb 26 09:35:28 2021

# @author: chris.kerklaan

# #TODO functions:
#     1. Memory counter

# Currently supported functions
#     1. Check properties
#         1. nodata
#         2. projection
#         3. data_type
#         4. extreme_values
#         5. max_allowed_pixels --> takes a long time, so currently switched off
#         6. Not square pixels --> could not create this hence not present
#     2. Check null rasters
#     3. Check load csv
#     4. Checks generate functions
#         1. Friction
#         2. Interceptions
#         3. Infiltration (Permeability and max infiltration)
#         4. Crop type
#         5. Hydrauilic conductivty
#         6. Building interception

# """
# # First-party imports
# import pathlib

# # Third-party imports
# import numpy as np
# from osgeo import gdal
# import logging

# logger = logging.getLogger()


# # Local imports
# from threedi_edits.threedi.rastergroup import ThreediRasterGroup
# from threedi_edits.gis.raster import Raster
# from threedi_edits.gis.vector import Vector

# # Globals
# TEST_DIRECTORY = (
#     str(pathlib.Path(__file__).parent.absolute()) + "/data/threedi_raster_group/"
# )

# MANDATORY_TITLES_SOIL = [
#     "Title",
#     "Code",
#     "Soiltype",
#     "Max_infiltration_rate",
#     "Hydraulic_conductivity",
# ]
# MANDATORY_TITLES_LANDUSE = [
#     "Code",
#     "Landuse",
#     "Friction",
#     "Permeability",
#     "Interception",
#     "Crop_type",
# ]


# table_path = TEST_DIRECTORY + "conversietabellen/"
# csv_landuse_file = table_path + "Conversietabel_landgebruik_2020.csv"
# csv_soil_file = table_path + "Conversietabel_bodem.csv"
# csv_landuse_file_broken = table_path + "Conversietabel_landgebruik_2020_broken.csv"
# dem_file = TEST_DIRECTORY + "dem.tif"
# interception_file = TEST_DIRECTORY + "interception.tif"
# frict_coef_file = TEST_DIRECTORY + "friction.tif"
# infiltration_file = TEST_DIRECTORY + "infiltration.tif"
# soil_file = TEST_DIRECTORY + "soil.tif"
# landuse_file = TEST_DIRECTORY + "landuse.tif"
# crop_type_file = TEST_DIRECTORY + "crop_type.tif"
# hydraulic_conductivity_file = TEST_DIRECTORY + "hydraulic_conductivity.tif"
# buildings_path = TEST_DIRECTORY + "buildings.shp"
# building_interception_file = TEST_DIRECTORY + "building_interception.tif"


# def test_check_nodata_properties():
#     """tests if checks work for nodata value equals -9999"""

#     group = ThreediRasterGroup(
#         dem_file=dem_file,
#         interception_file=interception_file,
#         frict_coef_file=frict_coef_file,
#         infiltration_rate_file=infiltration_file,
#     )
#     group.dem.nodata_value = -8888
#     checks = group.check_properties()

#     assert len(checks["errors"]) > 0

#     if len(checks["errors"]) > 0:
#         assert checks["errors"][0][0] == "nodata_value"


# def test_check_projection_properties():
#     """tests if checks work for projection and units if it is 28992 and metere"""

#     group = ThreediRasterGroup(
#         dem_file=dem_file,
#         interception_file=interception_file,
#         frict_coef_file=frict_coef_file,
#         infiltration_rate_file=infiltration_file,
#     )

#     group.dem.spatial_reference = 4326
#     checks = group.check_properties()

#     errors_present = len(checks["errors"]) > 0
#     assert errors_present
#     if errors_present:
#         assert checks["errors"][0][0] == "unit"
#         assert checks["errors"][1][0] == "projection"


# def test_check_data_type_properties():
#     """tests if checks work for data_type"""

#     group = ThreediRasterGroup(
#         dem_file=dem_file,
#         interception_file=interception_file,
#         frict_coef_file=frict_coef_file,
#         infiltration_rate_file=infiltration_file,
#     )
#     group.insert(group.dem.copy(shell=True, data_type=gdal.GDT_Int16))

#     checks = group.check_properties()

#     errors_present = len(checks["errors"]) > 0
#     assert errors_present

#     if errors_present:
#         assert checks["errors"][-1][0] == "data_type"


# def test_check_extreme_values_properties():
#     """tests if checks work for extreme values"""

#     group = ThreediRasterGroup(
#         dem_file=dem_file,
#         interception_file=interception_file,
#         frict_coef_file=frict_coef_file,
#         infiltration_rate_file=infiltration_file,
#     )
#     group.dem.load_to_memory()
#     group.dem.array = group.dem.array + 1 * 10**8
#     checks = group.check_properties()

#     errors_present = len(checks["errors"]) > 0
#     assert errors_present

#     if errors_present:
#         assert checks["errors"][1][0] == "extreme_values"


# def test_check_max_allowed_pixels_properties():
#     """tests if checks work for maximum allowed pixels by making an empty copy"""
#     group = ThreediRasterGroup(dem_file=dem_file)
#     copy = Raster.from_scratch(
#         "new",
#         group.dem.geotransform,
#         100000,
#         10001,
#     )
#     group["dem"] = copy
#     checks = group.check_properties()
#     errors_present = len(checks["errors"]) > 0
#     assert errors_present
#     assert checks["errors"][-1][0] == "maximum_allowed_pixels"
#     group = None


# def test_null_rasters():
#     """tests null rasters"""
#     group = ThreediRasterGroup(dem_file=dem_file)
#     array = group.null_raster().array
#     assert np.nansum(array) == 0.0


# def test_load_csv_soil():
#     """tests load csv on keys, values and if exists in object"""

#     group = ThreediRasterGroup(
#         dem_file=dem_file,
#         landuse_file=landuse_file,
#         soil_file=soil_file,
#     )

#     group.load_soil_conversion_table(csv_soil_file)
#     assert hasattr(group, "ct_soil")
#     if hasattr(group, "ct_soil"):
#         for title in MANDATORY_TITLES_SOIL:
#             assert title in group.ct_soil.keys()
#             values = group.ct_soil[title]
#             assert len(values) > 0


# def test_load_csv_lu():
#     """tests load csv on keys, values and if exists in object"""

#     group = ThreediRasterGroup(
#         dem_file=dem_file,
#         landuse_file=landuse_file,
#         soil_file=soil_file,
#     )

#     group.load_landuse_conversion_table(csv_landuse_file)
#     assert hasattr(group, "ct_lu")

#     if hasattr(group, "ct_lu"):
#         for title in MANDATORY_TITLES_LANDUSE:
#             assert title in group.ct_lu.keys()

#             values = group.ct_lu[title]
#             assert len(values) > 0


# def test_generate_friction():
#     """tests if generated friction using total counts"""

#     group = ThreediRasterGroup(dem_file=dem_file, landuse_file=landuse_file)

#     group.load_landuse_conversion_table(csv_landuse_file)
#     group.generate_friction()
#     or_sum = np.nansum(Raster(frict_coef_file).array)
#     generated_sum = np.nansum(group.friction.array)
#     assert or_sum == generated_sum


# def test_generate_interception():
#     """tests if generated interceptions using total counts"""

#     group = ThreediRasterGroup(dem_file=dem_file, landuse_file=landuse_file)

#     group.load_landuse_conversion_table(csv_landuse_file)
#     group.generate_interception()

#     or_sum = np.nansum(Raster(interception_file).array)
#     generated_sum = np.nansum(group.interception.array)
#     assert or_sum == generated_sum


# def test_generate_infiltration():
#     """tests if generated infiltraion using total counts"""

#     group = ThreediRasterGroup(
#         dem_file=dem_file,
#         landuse_file=landuse_file,
#         soil_file=soil_file,
#     )

#     group.load_landuse_conversion_table(csv_landuse_file)
#     group.load_soil_conversion_table(csv_soil_file)
#     group.generate_infiltration()

#     or_sum = np.nansum(Raster(infiltration_file).array)
#     generated_sum = np.nansum(group.infiltration.array)
#     assert or_sum == generated_sum


# def test_generate_crop_type():
#     """tests if generated crop type using total counts"""

#     group = ThreediRasterGroup(dem_file=dem_file, landuse_file=landuse_file)

#     group.load_landuse_conversion_table(csv_landuse_file)
#     group.generate_crop_type()

#     or_sum = np.nansum(Raster(crop_type_file).array)
#     generated_sum = np.nansum(group.crop_type.array)
#     assert or_sum == generated_sum


# def test_generate_hydraulic_conductivity():
#     """tests if generated hydraulic conductivity using total counts"""

#     group = ThreediRasterGroup(dem_file=dem_file, soil_file=soil_file)

#     group.load_soil_conversion_table(csv_soil_file)
#     group.generate_hydraulic_conductivity()
#     or_sum = np.nansum(Raster(hydraulic_conductivity_file).array)
#     generated_sum = np.nansum(group.hydraulic_conductivity.array)
#     assert or_sum == generated_sum


# def test_generate_building_interception():
#     """tests if generated buildings interception using total counts"""

#     group = ThreediRasterGroup(dem_file, buildings_file=buildings_path)
#     group.generate_building_interception(10)
#     group.interception.write(TEST_DIRECTORY + "building_interception2.tif")
#     or_sum = np.nansum(Raster(building_interception_file).array)
#     generated_sum = np.nansum(group.interception.array)
#     assert or_sum == generated_sum


# def test_generate_building_interception_nodata():
#     """tests if generated buildings interception using 0"""

#     group = ThreediRasterGroup(dem_file, buildings_file=buildings_path)
#     group.generate_building_interception(-9999)

#     assert group.interception.read(Vector(buildings_path)[2].geometry)[0] == 0


# def test_correct():
#     """tests if property correcting is properly perfomed"""
#     group = ThreediRasterGroup(
#         dem_file=dem_file,
#         interception_file=interception_file,
#         frict_coef_file=frict_coef_file,
#         infiltration_rate_file=infiltration_file,
#     )

#     group.interception.nodata_value = -8888
#     group.friction.spatial_reference = 28991
#     group.infiltration.change_data_type(int)

#     buildings = Vector(buildings_path)
#     group["infiltration"] = group.infiltration.push_vector(
#         buildings, set_to_nodata=True
#     )

#     checks = group.check_properties()
#     assert checks["errors"][1][0] == "nodata_value"
#     assert checks["errors"][2][0] == "projection"
#     assert checks["errors"][3][0] == "data_type"

#     group.correct()
#     checks = group.check_properties()

#     assert len(checks["errors"]) == 1 and checks["errors"][0][0] == "nodata_value"


# def test_write():
#     """tests write based on a dictionary"""
#     group = ThreediRasterGroup(
#         dem_file=dem_file,
#         interception_file=interception_file,
#         frict_coef_file=frict_coef_file,
#         infiltration_rate_file=infiltration_file,
#     )
#     group.write(TEST_DIRECTORY)


# def test_write_w_dict():
#     """tests write based on a dictionary"""
#     group = ThreediRasterGroup(
#         dem_file=dem_file,
#         interception_file=interception_file,
#         frict_coef_file=frict_coef_file,
#         infiltration_rate_file=infiltration_file,
#     )
#     group.write(TEST_DIRECTORY, output_names={"dem": "raster/dem.tif"})


# def test_clip():
#     """tests if the whole group is clipped"""
#     group = ThreediRasterGroup(
#         dem_file=dem_file,
#         interception_file=interception_file,
#         frict_coef_file=frict_coef_file,
#         infiltration_rate_file=infiltration_file,
#     )
#     group_sum = np.nansum(group.dem.array)
#     buildings = Vector(buildings_path)
#     group.clip(buildings)

#     assert group_sum < np.nansum(group.dem.array)


# def test_fix_table():
#     """tests if a broken csv is still loaded correctly
#     checks only if the lengths are still the same
#     """
#     group = ThreediRasterGroup()

#     group.load_landuse_conversion_table(csv_landuse_file_broken)
#     group.fix_table()

#     for c, (key, value) in enumerate(group.ct_lu.items()):

#         if c == 0:
#             base_len = len(value)
#         else:
#             if not key == "Landuse":
#                 assert len(value) == base_len


# # def test_memory_capacity():
# #     """ rastergroup should be able to handle three rasters of 500.000.000
# #         pixels each
# #     """
# #     group = ThreediRasterGroup(dem_file=dem_file,
# #                                 landuse_file=landuse_file,
# #                                 soil_file=soil_file)
# #     copy = group.dem.empty_copy(20000, 25000)
# #     group["dem"] = copy

# #     landuse = copy.empty_copy()
# #     for tile in copy:
# #         array = tile.array
# #         array[array==0]=1
# #         landuse.array = array, *tile.location
# #     landuse.name ="landuse"

# #     group["landuse"] = landuse

# #     soil = copy.empty_copy()
# #     for tile in copy:
# #         array = tile.array
# #         array[array==0]=1
# #         soil.array = array, *tile.location
# #     soil.name = "soil"
# #     group["soil"] = soil


# #     group.load_landuse_conversion_table(csv_landuse_file)
# #     group.load_soil_conversion_table(csv_soil_file)

# #     group.generate_friction()
# #     group.generate_permeability()
# #     group.generate_interception()
# #     group.generate_crop_type()
# #     group.generate_max_infiltration()
# #     group.generate_infiltration()
# #     group.generate_hydraulic_conductivity()

# #     assert True
# # if __name__ == "__main__":
# #     __file__ = "C:/Users/chris.kerklaan/Documents/Github/threedi-raster-edits/threedi_edits/tests/test_threedi_rastergroup.py"
# #     import sys

# #     sys.path.append("C:/Users/chris.kerklaan/Documents/Github/threedi-raster-edits")



# -*- coding: utf-8 -*-
"""
Created on Thu Nov 24 14:20:09 2022

@author: chris.kerklaan
"""

# versioning
__version__ = "1.1"

# Globals
SUPPORTED_THREEDI_VERSIONS = ["0209"]
INSTALL_REQUIRES = ["scipy", "cached_property", "shapely==2.0a1", "threedigrid_builder"]



# -*- coding: utf-8 -*-
"""
Created on Wed Dec 15 14:08:57 2021

@author: chris.kerklaan
"""
import os
import sys

sys.path.append(r"C:\Users\chris.kerklaan\Documents\Github\threedi-raster-edits")
from threedi_edits import Vector

os.chdir(r"C:\Users\chris.kerklaan\Documents\Projecten\basis kaarten hhnk")


def create(hydrolines, hydroparts, output_path):
    # first we write processing data
    # hydrolines.write("processing/tile/hydrolines.shp")
    # hydroparts.write("processing/tile/hydroparts.shp")

    # first buffer the lines
    hydrolines_buffered = hydrolines.buffer(1, quiet=False)
    # hydrolines_buffered.write("processing/tile/hydrolines_buffered.shp")

    # Then only select hydroparts which intersect the hydrolines
    hydroparts_filtered = hydroparts.spatial_filter(
        hydrolines, quiet=False, return_vector=True
    )
    # hydroparts_filtered.write("processing/tile/hydroparts_filtered.shp")

    # We merge to files
    merged = Vector.from_scratch("merged", 3, 28992)
    for hydroline in hydrolines_buffered:
        merged.add(hydroline)

    for hydropart in hydroparts_filtered:
        merged.add(hydropart)

    fixed = merged.fix(quiet=False)
    single = fixed.to_single(quiet=False)
    dissolved = single.dissolve(quiet=False)
    # dissolved.write("processing/tile/dissolved.shp")

    # We have to do vonoroi (influence area) of a linestring.
    # hence we make to points on a line, 10 cm from the start
    # and 10 cm from the end of the linestring to make sure
    # intersection work well.
    # Finally add some points halfway to ensure proper splits between adjacent
    # lines.
    voronoi_points = Vector.from_scratch("points", 1, 28992)
    voronoi_points.add_field("line_id", int)
    for hydroline in hydrolines:
        points_to_add = []
        for point in hydroline.geometry.points_on_line(
            custom_interval=[0.25, hydroline.geometry.length - 0.25],
            start=False,
            end=False,
        ):
            points_to_add.append(point)

        for point in hydroline.geometry.points_on_line(
            interval=5,
            start=False,
            end=False,
        ):
            points_to_add.append(point)

        for point in points_to_add:
            voronoi_points.add(geometry=point, items={"line_id": hydroline.fid})

    # voronoi_points.write("processing/tile/vonoroi_points.shp")

    # Create vonoroi polygons and dissolve the result
    # due to the fact that there are multiple points per line.
    vonoroi = voronoi_points.vonoroi()
    vonoroi = vonoroi.dissolve(field="line_id", quiet=False)
    # vonoroi.write("processing/tile/vonoroi4.shp")

    # Create the final vector by clipping it with the dissolved hydroparts.
    # We have to fix due to a 3D geometry, annoying, cannot be handled.
    singles = dissolved.to_single(quiet=False)
    singles_fixed = singles.fix(quiet=False)

    output = Vector.from_scratch("hydroparts", 3, 28992)
    output.add_field("line_id", int)

    for shape in vonoroi:
        for single in singles_fixed.spatial_filter(shape.geometry):
            clipped = single.geometry.clip(shape.geometry)
            if not clipped:
                print("Skipping shape", shape.id)
                continue
            output.add(geometry=clipped[0], items=shape.items)

    output.write(output_path, index=False)


hydrolines_path = "processing/data/hydroobjecten.shp"
hydroparts_path = "processing/data/bgt_waterdelen2.shp"
tiles_path = "processing/data/ahn_hhnk_grenzen_minor.shp"
tiles_major_path = "processing/data/AHN_HHNK_grenzen.shp"


hydrolines = Vector(hydrolines_path, create_index=True)
hydroparts = Vector(hydroparts_path, create_index=True)
tiles = Vector(tiles_path)
tiles_reversed = reversed(tiles)

for i, tile in enumerate(tiles_reversed):

    print(i, len(tiles), tile["name"])
    path = f"processing/tiles_minor/waterdeel_{tile['name']}.shp"
    linepath = f"processing/tiles_minor/hydrolijn_{tile['name']}.shp"
    if os.path.exists(path):
        print(path, "exists")
        continue

    if os.path.exists(linepath):
        tile_hydrolines = Vector(linepath)
    else:
        tile_hydrolines = hydrolines.spatial_filter(
            tile.geometry, return_vector=True, quiet=False, method="Extent"
        )

    if len(tile_hydrolines) == 0:
        if not os.path.exists(linepath):
            tile_hydrolines.write(linepath, overwrite=True, index=False)
        tile_hydrolines.write(path, overwrite=True, index=False)  # as a dummy
        continue

    tile_hydroparts = hydroparts.spatial_filter(
        tile.geometry, return_vector=True, quiet=False, method="Extent"
    )

    create(tile_hydrolines, tile_hydroparts, path)
    # tile_hydrolines.write(linepath, index=False)


# # # minor to major
# minor_tiles = Vector(tiles_path, create_table=True)
# major_tiles = Vector(tiles_major_path)

# for tile in major_tiles:
#     print(tile['name'])
#     waterdeel_path = f"processing/tiles/waterdeel_{tile['name']}.shp"
#     if os.path.exists(waterdeel_path):
#         print(f"{tile['name']} exists")
#         continue

#     skip = False
#     minors = []
#     for minor_tile in minor_tiles:
#         if tile['name'] in minor_tile['name']:
#             path = f"processing/tiles_minor/waterdeel_{minor_tile['name']}.shp"
#             if not os.path.exists(path):
#                 skip=True
#                 break
#             else:
#                 minors.append(minor_tile )

#     if skip == True:
#         print(f"Minor tile does not exist, skipping")
#         continue

#     output = Vector.from_scratch('merged', 3, 28992)
#     output.add_field("line_id", float)
#     for minor_tile in minors:
#         print("Adding", minor_tile['name'])
#         path = f"processing/tiles_minor/waterdeel_{minor_tile['name']}.shp"
#         minor = Vector(path)
#         minor = minor.fix(quiet=False)

#         minor_tile_vector = Vector.from_scratch("clip", 3, 28992)
#         minor_tile_vector.add(geometry=minor_tile.geometry.Buffer(0))
#         clipped = minor.clip(minor_tile_vector, quiet=False)
#         for feature in clipped:
#             output.add(feature)

#     print(len(output))
#     singles= output.to_single(quiet=False)
#     dissolved = singles.dissolve(field="line_id", quiet=False)

#     dissolved.write(waterdeel_path, index=False)


# # write hydrolines
# tiles_major_path = "processing/data/AHN_HHNK_grenzen.shp"
# hydrolines_path = "processing/data/hydroobjecten.shp"
# tiles = Vector(tiles_major_path)
# hydrolines = Vector(hydrolines_path)
# for i, tile in enumerate(tiles):
#     print(i, tile['name'])
#     linepath = f"processing/tiles/hydrolijn_{tile['name']}.shp"
#     tile_vector = Vector.from_scratch("clip", 3, 28992)
#     tile_vector.add(geometry=tile.geometry.Buffer(0))
#     hydrolines.set_spatial_filter(tile.geometry.Buffer(0))
#     tile_hydrolines = hydrolines.clip(tile_vector, quiet=False)
#     hydrolines.set_spatial_filter(None)

#     tile_hydrolines.write(linepath, index=False)

# merge all tiles into one

tiles = Vector(tiles_path)

output = Vector.from_scratch("all", 3, 28992)
output.add_field("line_id", int)

for tile in tiles:
    print(tile["name"])
    tile_vector = Vector.from_scratch("clip", 3, 28992)
    tile_vector.add(geometry=tile.geometry.Buffer(0))

    path = f"processing/tiles_minor/waterdeel_{tile['name']}.shp"
    waterdelen = Vector(path)
    waterdelen = waterdelen.fix(quiet=False)
    tile_hydroparts = waterdelen.clip(tile_vector, quiet=False, use_ogr=False)
    for waterdeel in tile_hydroparts:
        output.add(waterdeel)


singles = output.to_single(quiet=False)
fixed = singles.fix(quiet=False)
dissolved = fixed.dissolve(field="line_id", quiet=False)
dissolved.write("processing/processed_waterdeel4.shp")



# -*- coding: utf-8 -*-
"""
Created on Thu Feb 17 15:34:04 2022

@author: chris.kerklaan

@docs: Command line tool used for the downloads and conversion of 3Di rasters on a linux server.

"""
# find the package
import sys

sys.path.append(r"C:\Users\chris.kerklaan\Documents\Github\threedi-raster-edits")

# First imports
import pathlib
import threedi_edits as tre
import argparse

# Globals
CELLSIZE = 0.5
THREADS = 1
FIELD_NAME = None
AHN_VERSION = "ahn4"


def rextract_convert(
    vector_path: str,
    lizard_api_key: str,
    output_directory,
    field_name: str,
    cellsize,
    threads: int,
    friction: bool,
    infiltration: bool,
    interception: bool,
    landuse: bool,
    soil: bool,
    ahn: str,
):

    # create a dict
    threedi_uuids = {}
    if landuse:
        threedi_uuids["landuse"] = tre.UUID.THREEDI_LANDUSE

    if soil:
        threedi_uuids["soil"] = tre.UUID.THREEDI_SOIL

    get_ahn = {
        "ahn4": tre.UUID.THREEDI_AHN4,
        "ahn3": tre.UUID.THREEDI_AHN3,
        "ahn4_ahn3_combined": "6a75a859-e644-4498-a4c6-dbf367464ffd",
    }

    # select correct version
    threedi_uuids["ahn"] = get_ahn[ahn.lower()]

    # load you vector
    vector = tre.Vector(vector_path)

    # Create a rextract object
    rextract = tre.RasterExtraction(lizard_api_key)

    if field_name:
        if not field_name in vector.fields:
            raise ValueError(f"Field name {field_name} not found in {vector.fields}")

    for feature in vector:

        # retrieve the correct field name
        if field_name:
            output_folder_name = feature[field_name]
        else:
            output_folder_name = f"feature_{feature.fid}"

        # check if the folder exists, create otherwise
        folder = pathlib.Path(output_directory + "/" + output_folder_name)
        folder.mkdir(parents=True, exist_ok=True)

        # do download, store locations
        threedi_downloads = {}
        for name, threedi_uuid in threedi_uuids.items():
            path = str(folder / f"{name}.tif")
            rextract.run(
                path, threedi_uuid, feature.geometry, cellsize=cellsize, threads=threads
            )
            threedi_downloads[name] = path

        # do conversion
        if landuse and soil:
            group = tre.ThreediRasterGroup(
                # dem_file=threedi_downloads["ahn"],
                landuse_file=threedi_downloads["landuse"],
                soil_file=threedi_downloads["soil"],
                memory=False,
            )

            # load table
            group.load_landuse_conversion_table()
            group.load_soil_conversion_table()

            if infiltration:
                group.generate_infiltration()  # soil, landuse
                group.infiltration.spatial_reference = 28992  # TODO
                group.infiltration.write(str(folder / "infiltration.tif"))

            if friction:
                group.generate_friction()  # landuse
                group.friction.spatial_reference = 28992  # TODO
                group.friction.write(str(folder / "friction.tif"))

            if interception:
                group.generate_interception()  # landuse
                group.interception.spatial_reference = 28992  # TODO
                group.interception.write(str(folder / "interception.tif"))


def get_parser():
    """Return argument parser."""

    parser = argparse.ArgumentParser(description=__doc__)

    # main
    parser.add_argument("vector_path", help="Features of this vector are downloaded.")
    parser.add_argument("lizard_api_key", help="Lizard api key.")
    parser.add_argument("output_directory", help="Output rasters are placed here.")

    # options
    parser.add_argument(
        "-f", "--friction", action="store_true", help="Generates friction."
    )
    parser.add_argument(
        "-ic", "--interception", action="store_true", help="Generates interception"
    )
    parser.add_argument(
        "-if", "--infiltration", action="store_true", help="Generates infiltration"
    )
    parser.add_argument("-soil", "--soil", action="store_true", help="Downloads soil")
    parser.add_argument(
        "-landuse", "--landuse", action="store_true", help="Downloads landuse"
    )
    parser.add_argument(
        "-fn",
        "--field_name",
        default=FIELD_NAME,
        help="Field name used to place the rasters in a directory.",
    )
    parser.add_argument(
        "-cs", "--cellsize", default=CELLSIZE, type=float, help="Cellsize."
    )
    parser.add_argument(
        "-tr",
        "--threads",
        default=THREADS,
        type=int,
        help="Amount of threads used. More threads increases speed. Cannot be above 4.",
    )

    parser.add_argument(
        "-ahn",
        "--ahn",
        default=AHN_VERSION,
        type=str,
        help="AHN version: ahn4, ahn3 or ahn4_ahn3_combined",
    )

    parser.set_defaults(friction=False)
    parser.set_defaults(interception=False)
    parser.set_defaults(infiltration=False)
    parser.set_defaults(soil=False)
    parser.set_defaults(landuse=False)

    return parser


def main():
    """Call extract_all with args from parser."""
    return rextract_convert(**vars(get_parser().parse_args()))


if __name__ == "__main__":
    exit(main())



"""  
threedi_edits is a python toolbox for gis and threedi related processing.
Note that the only depedency for base processing should be gdal.

Release info for gdal:
https://github.com/OSGeo/gdal/blob/v3.4.1/gdal/NEWS.md

General imports from threedi_edits are listed below:
    
E.g., from threedi_edits import Raster

-- Use help(Raster) for further information
-- Use .modules for all possible imports of modules
-- Use .classes for all possible imports of classes
-- Use .functions for all possible imports of functions


gis-processing:
    - Raster
    - RasterGroup
    - Vector
    - VectorGroup
    - LineString
    - MultiLineString
    - Polygon
    - MultiPolygon
    - Point 
    - MultiPoint
    
threedi-processing
    - ThreediRasterGroup (used for the rasters of threedi)
    - ThreediEdits (used for the model database (.sqlite) of threedi)
            
Philosophy:
    We want to create a simple api on a threedi database.
    Using as few dependencies as possible.
    
About version 1.0:
    In version 1.0 we changed vector processing is changed by using shapely.
    
    
"""

from . import utils  # noqa
from .utils.logging import show_console_logging

show_console_logging()

from .utils.dependencies import DEPENDENCIES

DEPENDENCIES.missing()


from osgeo import gdal, ogr

# options
gdal.SetConfigOption("SQLITE_LIST_ALL_TABLES", "YES")
gdal.SetConfigOption("PG_LIST_ALL_TABLES", "YES")

# Exceptions
gdal.SetConfigOption("CPL_LOG", "NUl")
ogr.UseExceptions()
gdal.UseExceptions()


# gis
import shapely  # noqa
from . import gis  # noqa
from .gis.raster import Raster  # noqa
from .gis.rastergroup import RasterGroup  # noqa
from .gis.vector import Vector  # noqa
from .gis.vectorbase import VectorBase  # noqa
from .gis.vectorgroup import VectorGroup  # noqa
from .gis.linestring import LineString, MultiLineString  # noqa
from .gis.polygon import Polygon, MultiPolygon  # noqa
from .gis.point import Point, MultiPoint  # noqa

# Threedi
from . import threedi  # noqa
from .threedi.rastergroup import ThreediRasterGroup  # noqa
from .threedi.rastergroup import retrieve_soil_conversion_table  # noqa
from .threedi.rastergroup import retrieve_landuse_conversion_table  # noqa
from .threedi.edits import ThreediEdits  # noqa

# Logging
from . import utils  # noqa
from .utils.progress import Progress  # noqa
from .utils.project import log_time  # noqa
from .utils.project import Files  # noqa
from .utils.project import Functions  # noqa
from .utils.project import Classes  # noqa
from .utils.project import Modules  # noqa
from .utils.project import Logger  # noqa

# Examples
from .examples import get_examples  # noqa

# Globals
from .globals import SUPPORTED_THREEDI_VERSIONS  # noqa
from .globals import INSTALL_REQUIRES  # noqa
from .globals import __version__  # noqa


# structure
files = Files(__file__)
functions = Functions(__name__)
classes = Classes(__name__, local_only=False)
modules = Modules(__name__)



# -*- coding: utf-8 -*-
"""
Created on Tue Dec 14 10:39:47 2021

@author: chris.kerklaan

Steps for sampling/reading data from vector geometries and burning it back to 
the dem. 
"""
import os
import numpy as np
from threedi_edits import Vector, Raster, Progress, RasterGroup

# dem_path = r"data/09bn2.tif"
# buildings_path = r"data/Panden_nieuw.shp"
# tegel_path = r"data/AHN_HHNK_grenzen.shp"

# os.chdir(r"C:\Users\chris.kerklaan\Documents\Projecten\basis kaarten hhnk\processing")

# # load raster
# dem = Raster(dem_path)

# # get dem tile
# tegels = Vector(tegel_path)
# tegel = tegels.filter(name="09bn2")[0]

# # get buildings within tile.
# buildings = Vector(buildings_path, create_index=True)
# buildings_within_tile = buildings.spatial_filter(tegel.geometry)

# # make output
# output = buildings.copy(shell=True)
# output.add_field("level", float)

# # find percentile, add buffer
# for building in buildings_within_tile:

#     # do a 1 meter buffer
#     geometry_buffered = building.geometry.buffer(1)

#     # difference with the original to create a hull
#     difference = geometry_buffered.difference(building.geometry)

#     # read the data
#     data = dem.read(difference)

#     # find the percentile
#     percentile = np.percentile(data, 75)

#     # percentile + 5 cm
#     floorheight = percentile + 0.05

#     # write to output with new fields
#     items = building.items
#     items["level"] = floorheight
#     output.add(items=items, geometry=building.geometry)

# output.write("building_output.shp")

# # find out how to find functions
# dir(dem)

# # find arguments
# help(dem.push_vector)

# # push the data onto the raster
# dem_floorheight = dem.push_vector(output, field="level")
# dem_floorheight.write("dem_level.tif")


# # Use the above data as a function a loop over multiple tiles
# def burn_floorlevels_in_tile(dem_tile: Raster, buildings: Vector, output_raster_path):
#     # get buildings within tile. Slightly different then the above method
#     # we are now directory getting the extent from the raster with raster.extent_geometry
#     buildings_within_tile = buildings.spatial_filter(dem_tile.extent_geometry)

#     # make output
#     output = buildings.copy(shell=True)
#     output.add_field("level", float)

#     # find percentile, add buffer
#     for building in buildings_within_tile:

#         # do a 1 meter buffer
#         geometry_buffered = building.geometry.buffer(1)

#         # difference with the original to create a hull
#         difference = geometry_buffered.difference(building.geometry)

#         # read the data
#         data = dem_tile.read(difference)

#         # find the percentile
#         percentile = np.percentile(data, 75)

#         # percentile + 5 cm
#         floorheight = percentile + 0.05

#         # write to output with new fields
#         items = building.items
#         items["level"] = floorheight
#         output.add(items=items, geometry=building.geometry)

#     dem_floorheight = dem_tile.push_vector(output, field="level")
#     dem_floorheight.write(output_raster_path)


# # Now we are using this function to automate our processes
# dem_directory = r"\\utr-3fs-01.nens.local\WorkDir\J_vLieshout\HHNK_AHN_basiskaarten\1_AHN4_Interpolated\ahn4"

# # We can only show our progress using the Progress class as is show below
# for dem_file in Progress(os.listdir(dem_directory), "Generating floorlevels"):

#     # skip the file if it does not end with '.tif'
#     if not dem_file.endswith(".tif"):
#         continue

#     # load data
#     dem_tile = Raster(dem_directory + "/" + dem_file)
#     buildings = Vector(
#         r"\\utr-3fs-01.nens.local\WorkDir\J_vLieshout\HHNK_AHN_basiskaarten\Panden\panden_met_index.shp"
#     )

#     # make output path
#     output_raster_path = (
#         r"\\utr-3fs-01.nens.local\WorkDir\J_vLieshout\HHNK_AHN_basiskaarten\1_AHN4_Interpolated\ahn4_vloerpijl/"
#         + dem_file
#     )

#     # start processing
#     burn_floorlevels_in_tile(dem_tile, buildings, output_raster_path)


## bruggen

panden_path = r"C:\Users\chris.kerklaan\Documents\Projecten\basis kaarten hhnk\processing\data\Panden_nieuw.shp"
buildings = Vector(panden_path)
dem_tiles_dir = r"K:\J_vLieshout\HHNK_AHN_basiskaarten\AHN4_DTM"

dem_tiles = [
    Raster(dem_tiles_dir + f"/{i}")
    for i in os.listdir(dem_tiles_dir)
    if i.endswith("tif")
]
dem = RasterGroup(dem_tiles)

# buffered = buildings.buffer(1, quiet=False)

# buffered = Vector( r"C:\Users\chris.kerklaan\Documents\Projecten\basis kaarten hhnk\processing\data\panden_buffered.shp")
# clusters = buffered.dissolve(intersecting=True, quiet=False)
# clusters.write(r"C:\Users\chris.kerklaan\Documents\Projecten\basis kaarten hhnk\processing\data\panden_clusters.shp")


buildings = Vector(
    r"C:\Users\chris.kerklaan\Documents\Projecten\basis kaarten hhnk\processing\data\Panden_nieuw.shp"
)
clusters = Vector(
    r"C:\Users\chris.kerklaan\Documents\Projecten\basis kaarten hhnk\processing\data\panden_clusters.shp"
)

output = buildings.copy(shell=True)
for cluster in Progress(clusters, "Getting the sample areas"):

    building_ids = cluster["diss_ids"].split("_")
    for building_id in building_ids:

        if building_id == "":
            continue
        building = buildings[int(building_id)]

        # do a 1 meter buffer
        geometry_buffered = building.geometry.buffer(1)

        # difference with the original to create a hull
        difference = geometry_buffered.difference(building.geometry)

        # # also it cannot intersect with other builindgs
        for intersecting_building_id in building_ids:
            if intersecting_building_id == building.id:
                continue
            if intersecting_building_id == "":
                continue

            intersecting_building = buildings[int(intersecting_building_id)]

            intersecting_buffer = intersecting_building.geometry.buffer(1)
            difference = difference.difference(intersecting_buffer)

        if difference.IsEmpty():
            continue
        if difference.GetGeometryType() == 3:

            output.add(geometry=difference, fid=int(building_id))

output.write(
    r"C:\Users\chris.kerklaan\Documents\Projecten\basis kaarten hhnk\processing/data/panden_sample_area2.shp"
)

buildings = Vector(
    r"C:\Users\chris.kerklaan\Documents\Projecten\basis kaarten hhnk\processing\panden_sampled/Panden_nieuw.shp"
)
new = buildings.copy(shell=True)
new.add_field("vloer", float)
for building in Progress(buildings, "buildings"):
    new.add(
        items=building.items,
        vloer=building["level_75"] + 0.05,
        geometry=building.geometry,
    )


for i, dem_tile_name in enumerate(os.listdir(dem_tiles_dir)):
    if not dem_tile_name.endswith("tif"):
        continue

    output_path = rf"C:\Users\chris.kerklaan\Documents\Projecten\basis kaarten hhnk\processing\dtm_interpolated/{dem_tile_name}"
    if os.path.exists(output_path):
        print("Exists")
        continue

    print(dem_tile_name, f"{i}/{len(os.listdir(dem_tiles_dir))}")
    dem_tile = Raster(
        rf"G:\Projecten W (2021)\W0234 - Nieuwe basiskaarten 3Di voor HHNK\Gegevens\Resultaat\21-12-2021 - Oplevering\Geinterpoleerde DEM\{dem_tile_name}"
    )
    new_dem = dem_tile.push_vector(new, field="vloer")
    new_dem.write(output_path)


output = Vector.from_scratch("panden", 3, 28992)
output.add_field("fid", int)

output.add_field("gid", int)
output.add_field("bouwjaar", int)
output.add_field("identifica", str)
output.add_field("pandstatus", str)
output.add_field("vloerpeil_75", float)
output.add_field("vloerpeil_50", float)
output.add_field("vloerpeil_25", float)
output.add_field("vloerpeil_min", float)
output.add_field("vloerpeil_max", float)

for building in Progress(buildings, "Sampling panden"):

    # do a 1 meter buffer
    geometry_buffered = building.geometry.buffer(1)

    # difference with the original to create a hull
    difference = geometry_buffered.difference(building.geometry)

    # # also it cannot intersect with other builindgs
    for intersecting_building_id in building_ids:
        intersecting_building = buildings[int(intersecting_building_id)]
        if intersecting_building.id == building.id:
            continue
        intersecting_buffer = intersecting_building.geometry.buffer(1)
        difference = difference.difference(intersecting_buffer)

    # read the data
    data = dem.read(difference)

    if len(data) == 0:
        # do a 1 meter buffer
        geometry_buffered = building.geometry.buffer(3)

        # difference with the original to create a hull
        difference = geometry_buffered.difference(building.geometry)

        # read the data
        data = dem.read(difference)

    if len(data) == 0:
        data = [np.nan]

    # find the percentile
    percentile_75 = np.nanpercentile(data, 75) + 0.05
    percentile_50 = np.nanpercentile(data, 50) + 0.05
    percentile_25 = np.nanpercentile(data, 25) + 0.05
    percentile_min = np.nanmin(data) + 0.05
    percentile_max = np.nanmax(data) + 0.05

    # write to output with new fields
    items = building.items
    items["percentile_75"] = percentile_75
    items["percentile_50"] = percentile_50
    items["percentile_25"] = percentile_25
    items["percentile_min"] = percentile_min
    items["percentile_max"] = percentile_max

    output.add(items=items, geometry=building.geometry)
output.write(
    r"C:\Users\chris.kerklaan\Documents\Projecten\basis kaarten hhnk\processing\panden_sampled\panden_hhnk.gpkg"
)



# -*- coding: utf-8 -*-
"""
@author: chris.kerklaan

#TODO:

#Done:
    1. from_points
    2. __iter__
    3. get_points
    4. reversed
"""
# First-party imports
import pathlib

# Local imports
from threedi_edits.gis.linestring import LineString
from threedi_edits import Vector
from threedi_edits.gis.point import Point

# Globals
DATA = pathlib.Path(__file__).parent / "data" / "gis_linestring.gpkg"
line = Vector(DATA, layer_name="line")[0].geometry.ogr
split_point = Vector(DATA, layer_name="split_point")[0].geometry.ogr


def test_from_points():
    """tests if a geometry can be created from points"""
    points = [(1116651.4, 637392.6), (1188804.1, 652655.7)]
    line = LineString.from_points(points)
    assert type(line) == LineString


def test__iter__():
    """tests if ouput if iterations has count and is unequal"""
    line_geom = LineString(line)
    lines = [l for l in line_geom]
    assert len(lines) == 2
    assert lines[0].wkt != lines[1].wkt


def test_start_point():
    """tests if start point has original value"""
    line_geom = LineString(line)
    assert line_geom.start_point.point == (
        117565.43122004971,
        459721.3204805324,
    )


def test_end_point():
    """tests if start point has original value"""
    line_geom = LineString(line)
    assert line_geom.end_point.point == (
        117486.66500118251,
        459702.01357949915,
    )


def test_reversed():
    line_geom = LineString(line)
    reverse = line_geom.reversed()

    assert reverse.start_point.point == (
        117486.66500118251,
        459702.01357949915,
    )
    assert reverse.end_point.point == (117565.43122004971, 459721.3204805324)


def test_points_on_line_distance():
    line_geom = LineString(line)
    points = line_geom.points_on_line(10)
    assert int(points[3].Distance(points[4])) == 10


def test_points_on_line_wo_start_end():
    line_geom = LineString(line)
    points = line_geom.points_on_line(10, start=False, end=False)
    assert len(points) == 8


def test_points_on_line_w_vertices():
    line_geom = LineString(line)
    points = line_geom.points_on_line(10, vertices=True)
    assert len(points) == 11


def test_perendicular_lines():
    """tests if the lines have correct line length"""
    line_geom = LineString(line)
    perps = line_geom.perpendicular_lines(10, 1)

    assert int(perps[0].length) == 2
    assert len(perps) == 10


def test_add_vertice():
    """tests if extra vertice is added at point"""
    line_geom = LineString(line)
    extra = line_geom.add_vertice(Point(split_point))
    assert len(extra.points) == 4



# -*- coding: utf-8 -*-
"""
Created on Fri Apr 26 14:39:17 2019

@author: chris.kerklaan
"""
import pathlib
import shutil
from threedi_edits.utils.project import Files
from threedi_edits.utils.project import Modules

files = Files(__file__)
modules = Modules(__name__)


def get_examples(destination_dir):
    current_directory = str(pathlib.Path(__file__).parent)
    shutil.copytree(current_directory, destination_dir)
